
from lxml import etree as etree_
from .hostedservice import HostedServiceClient
from ..namespaces import msgTag
from ..pmtypes import LocalizedText

class LocalizationServiceClient(HostedServiceClient):

    def _getLocalizedTextResponse(self, refs=None, version=None, langs=None, textWidths=None, numberOfLines=None, request_manipulator=None):
        """

        :param refs: a list of strings or None
        :param version: an unsigned integer or None
        :param langs: a list of strings or None
        :param textWidths: a list of strings or None (each string one of xs, s, m, l, xs, xxs)
        :param numberOfLines: a list of unsigned integers or None
        :param request_manipulator:
        :return: a list of LocalizedText objects
        """
        requestparams = []
        if refs is not None:
            for r in refs:
                node = etree_.Element(msgTag('Ref'))
                node.text = r
                requestparams.append(node)
        if version is not None:
            node = etree_.Element(msgTag('Version'))
            node.text = str(version)
            requestparams.append(node)
        if langs is not None:
            for l in langs:
                node = etree_.Element(msgTag('Lang'))
                node.text = l
                requestparams.append(node)
        if textWidths is not None:
            for tw in textWidths:
                node = etree_.Element(msgTag('TextWidth'))
                node.text = tw
                requestparams.append(node)
        if numberOfLines is not None:
            for nol in numberOfLines:
                node = etree_.Element(msgTag('NumberOfLines'))
                node.text = nol
                requestparams.append(node)

        resultSoapEnvelope = self._callGetMethod('GetLocalizedText', params=requestparams,
                                                 request_manipulator=request_manipulator)
        return resultSoapEnvelope

    def getLocalizedTextNode(self, refs=None, version=None, langs=None, textWidths=None, numberOfLines=None, request_manipulator=None):
        return self._getLocalizedTextResponse(refs, version, langs, textWidths, numberOfLines, request_manipulator).msgNode

    def getLocalizedTexts(self, refs=None, version=None, langs=None, textWidths=None, numberOfLines=None, request_manipulator=None):
        result = []
        responseNode = self._getLocalizedTextResponse(refs, version, langs, textWidths, numberOfLines, request_manipulator).msgNode
        if responseNode is not None:
            for element in responseNode:
                lt = LocalizedText.fromNode(element)
                result.append(lt)
        return result

    def getSupportedLanguages(self, request_manipulator=None):
        resultSoapEnvelope = self._callGetMethod('GetSupportedLanguages', request_manipulator=request_manipulator)
        result = []
        for element in resultSoapEnvelope.msgNode:
            result.append(str(element.text))
        return result

    def getSupportedLanguagesNodes(self, request_manipulator=None):
        resultSoapEnvelope = self._callGetMethod('GetSupportedLanguages', request_manipulator=request_manipulator)
        return resultSoapEnvelope.msgNode




import time
import uuid
from collections import OrderedDict
from collections import namedtuple
from contextlib import contextmanager
from functools import wraps
from threading import Lock

from lxml import etree as etree_

from . import mdibbase
from . import msgreader
from .devicewaveform import DefaultWaveformSource
from .. import loghelper
from .. import pmtypes
from ..definitions_base import ProtocolsRegistry
from ..definitions_sdc import SDC_v1_Definitions
from ..namespaces import domTag
from ..namespaces import extTag

_TrItem = namedtuple('_TrItem', 'old new')  # a named tuple for better readability of code


class _TransactionBase(object):
    def __init__(self, device_mdib_container):
        self._deviceMdibContainer = device_mdib_container
        self.descriptorUpdates = OrderedDict()
        self.metricStateUpdates = OrderedDict()
        self.alertStateUpdates = OrderedDict()
        self.componentStateUpdates = OrderedDict()
        self.contextStateUpdates = OrderedDict()
        self.operationalStateUpdates = OrderedDict()
        self.rtSampleStateUpdates = dict()  # unordered dict for performance
        self._error = False
        self._closed = False
        self.mdib_version = None

    def _getDescriptorInTransaction(self, descriptorHandle):
        """ looks for new or updated descriptor in current transaction and in mdib"""
        tr_containers = self.descriptorUpdates.get(descriptorHandle)
        if tr_containers is not None:
            old, new = tr_containers
            if new is None:  # descriptor is deleted in this transaction!
                raise RuntimeError('The descriptor {} is going to be deleted'.format(descriptorHandle))
            else:
                return new
        else:
            return self._deviceMdibContainer.descriptions.handle.getOne(descriptorHandle)

    def _get_or_mk_StateContainer(self, descriptorHandle, adjustStateVersion=True):
        """ returns oldContainer, newContainer"""
        descriptorContainer = self._getDescriptorInTransaction(descriptorHandle)
        old_stateContainer = self._deviceMdibContainer.states.descriptorHandle.getOne(descriptorContainer.handle,
                                                                                      allowNone=True)
        if old_stateContainer is None:
            # create a new state object
            new_stateContainer = self._deviceMdibContainer.mkStateContainerFromDescriptor(descriptorContainer)
            if adjustStateVersion:
                self._deviceMdibContainer.states.setVersion(new_stateContainer)
        else:
            new_stateContainer = old_stateContainer.mkCopy()
            new_stateContainer.incrementState()
        return old_stateContainer, new_stateContainer


def tr_method_wrapper(method):
    """a decorator for consistency checks and error handling"""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self._closed:
            raise RuntimeError('This Transaction is closed!')
        if self._error:
            raise RuntimeError('This Transaction failed due to an previous error!')
        try:
            return method(self, *args, **kwargs)
        except:
            self._error = True
            raise

    return wrapper


class _RtDataMdibUpdateTransaction(_TransactionBase):
    """This transaction is only used internally to periodically send waveform notifications.
    It handles this specific purpose with less overhead compared to regular transaction."""

    def __init__(self, device_mdib_container):
        super().__init__(device_mdib_container)

    @tr_method_wrapper
    def getRealTimeSampleArrayMetricState(self, descriptorHandle):
        # for performance reasons, this method does not return a copy of the original object.
        # This means no rollback possible.
        if descriptorHandle in self.rtSampleStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        stateContainer = self._deviceMdibContainer.states.descriptorHandle.getOne(descriptorHandle, allowNone=True)
        if stateContainer is None:
            raise ValueError('state {} not found!'.format(descriptorHandle))

        if not stateContainer.isRealtimeSampleArrayMetricState:
            raise ValueError('descriptorHandle {} does not reference a RealTimeSampleArrayMetricState'.format(
                descriptorHandle))
        stateContainer.incrementState()
        new_state = stateContainer  # supply old and new state; although identical, just do not break interface
        self.rtSampleStateUpdates[descriptorHandle] = _TrItem(stateContainer, new_state)
        return new_state


class _MdibUpdateTransaction(_TransactionBase):
    # pylint: disable=protected-access
    """ a helper class that collects multiple updates into one transaction.
    Used by contextmanager DeviceMdibContainer.mdibUpdateTransaction """

    def __init__(self, device_mdib_container):
        super().__init__(device_mdib_container)

    @tr_method_wrapper
    def addDescriptor(self, descriptorContainer, adjustDescriptorVersion=True):
        """
        :param descriptorContainer: the object that shall be added to mdib
        :param adjustDescriptorVersion: if True, and a descriptor with this handle does not exist, but was already present in this mdib before,
          the DescriptorVersion of descriptorContainer is set to last known version for this handle +1
        :return: None
        """
        descriptorHandle = descriptorContainer.handle
        if adjustDescriptorVersion:
            self._deviceMdibContainer.descriptions.setVersion(descriptorContainer)
        if descriptorHandle in self.descriptorUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        if descriptorHandle in self._deviceMdibContainer.descriptions.handle.keys():
            raise ValueError('cannot create descriptorHandle {}, it already exists!'.format(descriptorHandle))
        self.descriptorUpdates[descriptorHandle] = _TrItem(None, descriptorContainer)

    createDescriptor = addDescriptor

    @tr_method_wrapper
    def removeDescriptor(self, descriptorHandle):
        if descriptorHandle in self.descriptorUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        origDescriptorContainer = self._deviceMdibContainer.descriptions.handle.getOne(descriptorHandle)
        self.descriptorUpdates[descriptorHandle] = _TrItem(origDescriptorContainer, None)

    @tr_method_wrapper
    def getDescriptor(self, descriptorHandle):
        """ Update a descriptor.
        When the transaction is committed, the modifications to the copy will be applied to the original version,
        and notification messages will be sent to clients.
        @return: a copy of the state.
        """
        if descriptorHandle in self.descriptorUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        origDescriptorContainer = self._deviceMdibContainer.descriptions.handle.getOne(descriptorHandle)
        descriptorContainer = origDescriptorContainer.mkCopy()
        descriptorContainer.incrementDescriptorVersion()
        self.descriptorUpdates[descriptorHandle] = _TrItem(origDescriptorContainer, descriptorContainer)
        return descriptorContainer

    def hasState(self, descriptorHandle):
        """ check if transaction has a state with given handle """
        return self.getStateTransactionItem(descriptorHandle) is not None

    def getStateTransactionItem(self, descriptorHandle):
        """ if transaction has a state with given handle, return the transaction-item, otherwise None.  """
        for lookup in (self.metricStateUpdates,
                       self.alertStateUpdates,
                       self.componentStateUpdates,
                       self.contextStateUpdates,
                       self.operationalStateUpdates,
                       self.rtSampleStateUpdates):
            if descriptorHandle in lookup:
                return lookup[descriptorHandle]

    @tr_method_wrapper
    def addState(self, stateContainer, adjustStateVersion=True):
        my_multi_key = self._deviceMdibContainer.states
        my_updates = []
        if stateContainer.isMetricState:
            my_updates = self.metricStateUpdates
        elif stateContainer.isSystemContextState:
            my_updates = self.metricStateUpdates
        elif stateContainer.isAlertState:
            my_updates = self.alertStateUpdates
        elif stateContainer.isComponentState:
            my_updates = self.componentStateUpdates
        elif stateContainer.isOperationalState:
            my_updates = self.operationalStateUpdates
        elif stateContainer.isContextState:
            my_updates = self.contextStateUpdates
            my_multi_key = self._deviceMdibContainer.contextStates
        elif stateContainer.isRealtimeSampleArrayMetricState:
            my_updates = self.rtSampleStateUpdates
        elif stateContainer.NODETYPE == domTag('ScoState'):
            # special case ScoState Draft6: cannot notify updates, it is a category of its own that does not fit anywhere
            # This is a bug in the spec, not in this implementation!
            return

        descriptorHandle = stateContainer.descriptorHandle
        my_handle = stateContainer.Handle if stateContainer.isContextState else stateContainer.descriptorHandle
        if my_handle in my_updates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))

        if adjustStateVersion:
            my_multi_key.setVersion(stateContainer)
        my_updates[descriptorHandle] = _TrItem(None, stateContainer)  # old, new

    def ungetState(self, stateContainer):
        """ forget a state that was provided before by a getXXXState call"""
        for lookup in (self.alertStateUpdates, self.componentStateUpdates, self.contextStateUpdates,
                       self.metricStateUpdates, self.operationalStateUpdates, self.rtSampleStateUpdates):
            if stateContainer.descriptorHandle in lookup:
                del lookup[stateContainer.descriptorHandle]

    @tr_method_wrapper
    def getMetricState(self, descriptorHandle, adjustStateVersion=True):
        """ Update a MetricState.
        When the transaction is committed, the modifications to the copy will be applied to the original version,
        and notification messages will be sent to clients.
        :param descriptorHandle: the descriptorHandle of the object that shall be read
        :param adjustStateVersion: if True, and a state with this handle does not exist, but was already present in this mdib before,
          the StateVersion of descriptorContainer is set to last known version for this handle +1
        @return: a copy of the state.
        """
        if descriptorHandle in self.metricStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        old_state, new_state = self._get_or_mk_StateContainer(descriptorHandle, adjustStateVersion)
        if not new_state.isMetricState:
            raise ValueError('descriptorHandle {} does not reference a metric state'.format(descriptorHandle))
        self.metricStateUpdates[descriptorHandle] = _TrItem(old_state, new_state)
        return new_state

    @tr_method_wrapper
    def getComponentState(self, descriptorHandle, adjustStateVersion=True):
        """ Update a ComponentState.
        When the transaction is committed, the modifications to the copy will be applied to the original version,
        and notification messages will be sent to clients.
        :param descriptorHandle: the descriptorHandle of the object that shall be read
        :param adjustStateVersion: if True, and a state with this handle does not exist, but was already present in this mdib before,
          the StateVersion of descriptorContainer is set to last known version for this handle +1
        @return: a copy of the state.
        """
        if descriptorHandle in self.componentStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        old_state, new_state = self._get_or_mk_StateContainer(descriptorHandle, adjustStateVersion)
        if not new_state.isComponentState:
            raise ValueError('descriptorHandle {} does not reference a component state'.format(descriptorHandle))
        self.componentStateUpdates[descriptorHandle] = _TrItem(old_state, new_state)
        return new_state

    @tr_method_wrapper
    def getAlertState(self, descriptorHandle, adjustStateVersion=True):
        """ Update AlertConditionState or AlertSignalState node
        When the transaction is committed, the modifications to the copy will be applied to the original version,
        and notification messages will be sent to clients.
        :param descriptorHandle: the descriptorHandle of the object that shall be read
        :param adjustStateVersion: if True, and a state with this handle does not exist, but was already present
          in this mdib before,
          the StateVersion of descriptorContainer is set to last known version for this handle +1
        @return: a copy of the state.
        """
        if descriptorHandle in self.alertStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        old_state, new_state = self._get_or_mk_StateContainer(descriptorHandle, adjustStateVersion)
        if not new_state.isAlertState:
            raise ValueError('descriptorHandle {} does not reference an alert state'.format(descriptorHandle))
        self.alertStateUpdates[descriptorHandle] = _TrItem(old_state, new_state)
        return new_state

    @tr_method_wrapper
    def getContextState(self, descriptorHandle, contextStateHandle=None, adjustStateVersion=True, set_associated=True):
        """ Create or update a ContextState.

        If contextStateHandle is None or the handle does not exist in mdib, a new Context State
        is created and returned. Else an existing ContextState with that handle is returned.
        :param descriptorHandle: the descriptorHandle of the ContextState that shall be read.
        :param contextStateHandle: the handle of the ContextState that shall be read.
        :param adjustStateVersion: if True, and a state with this handle does not exist,
          but was already present in this mdib before, the StateVersion of descriptorContainer is set to
          last known version for this handle +1
        :param set_associated: If True, and a new context state is created, ContextAssociation is set to 'Assoc'
          and BindingMdibVersion + BindingStartTime are set.
        @return: a copy of the state.
        """
        lookup_key = (descriptorHandle, contextStateHandle)
        if lookup_key in self.contextStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(lookup_key))
        descriptorContainer = self._getDescriptorInTransaction(descriptorHandle)
        if contextStateHandle is None:
            # create a new context state with random handle.
            oldStateContainer = None
            newStateContainer = self._deviceMdibContainer.mkStateContainerFromDescriptor(descriptorContainer)
        else:
            oldStateContainer = self._deviceMdibContainer.contextStates.handle.getOne(contextStateHandle,
                                                                                      allowNone=True)
            if oldStateContainer is not None:
                if oldStateContainer.descriptorHandle != descriptorHandle:
                    raise ValueError(
                        f'Handle {contextStateHandle} and DescriptorHandle {descriptorHandle} do not match.')
                set_associated = False  # keep state as it is
                newStateContainer = oldStateContainer.mkCopy()
                newStateContainer.incrementState()
            else:
                # create a new context state with given handle.
                newStateContainer = self._deviceMdibContainer.mkStateContainerFromDescriptor(descriptorContainer)
                newStateContainer.Handle = contextStateHandle
                if adjustStateVersion:
                    self._deviceMdibContainer.contextStates.setVersion(newStateContainer)
        if set_associated:
            newStateContainer.BindingMdibVersion = self._deviceMdibContainer.mdibVersion + 1  # version after commit
            newStateContainer.BindingStartTime = time.time()
            newStateContainer.ContextAssociation = pmtypes.ContextAssociation.ASSOCIATED
        self.contextStateUpdates[lookup_key] = _TrItem(oldStateContainer, newStateContainer)
        return newStateContainer

    @tr_method_wrapper
    def addContextState(self, contextStateContainer, adjustStateVersion=True):
        """ Add a new ContextState.
        :param contextStateContainer: a ContextStateContainer instance
        :param adjustStateVersion: if True, and a state with this handle does not exist,
          but was already present in this mdib before,
          the StateVersion of descriptorContainer is set to last known version for this handle +1
        """
        lookup_key = (contextStateContainer.descriptorHandle, contextStateContainer.Handle)
        if lookup_key in self.contextStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(lookup_key))
        if adjustStateVersion:
            self._deviceMdibContainer.contextStates.setVersion(contextStateContainer)
        self.contextStateUpdates[lookup_key] = _TrItem(None, contextStateContainer)

    @tr_method_wrapper
    def getOperationalState(self, descriptorHandle, adjustStateVersion=True):
        """ Update an OperationalState.
        When the transaction is committed, the modifications to the copy will be applied to the original version,
        and notification messages will be sent to clients.
        :param descriptorHandle: the descriptorHandle of the object that shall be read
        :param adjustStateVersion: if True, and a state with this handle does not exist, but was already present in this mdib before,
          the StateVersion of descriptorContainer is set to last known version for this handle +1
        @return: a copy of the state.
        """
        if descriptorHandle in self.operationalStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        old_state, new_state = self._get_or_mk_StateContainer(descriptorHandle, adjustStateVersion)
        if not new_state.isOperationalState:
            raise ValueError('descriptorHandle {} does not reference an operational state '
                             '({})'.format(descriptorHandle, new_state.__class__.__name__))
        self.operationalStateUpdates[descriptorHandle] = _TrItem(old_state, new_state)
        return new_state

    @tr_method_wrapper
    def getRealTimeSampleArrayMetricState(self, descriptorHandle):
        if descriptorHandle in self.rtSampleStateUpdates:
            raise ValueError('descriptorHandle {} already in updated set!'.format(descriptorHandle))
        state_container = self._deviceMdibContainer.states.descriptorHandle.getOne(descriptorHandle, allowNone=True)
        if state_container is None:
            descriptorContainer = self._getDescriptorInTransaction(descriptorHandle)
            new_state = self._deviceMdibContainer.mkStateContainerFromDescriptor(descriptorContainer)
            if not new_state.isRealtimeSampleArrayMetricState:
                raise ValueError(
                    'descriptorHandle {} does not reference a RealTimeSampleArrayMetricState'.format(descriptorHandle))
            self._deviceMdibContainer.states.addObject(state_container)
        else:
            if not state_container.isRealtimeSampleArrayMetricState:
                raise ValueError(
                    'descriptorHandle {} does not reference a RealTimeSampleArrayMetricState'.format(descriptorHandle))
            new_state = state_container.mkCopy(copy_node=False)
            new_state.incrementState()
        self.rtSampleStateUpdates[descriptorHandle] = _TrItem(state_container, new_state)
        return new_state


class DeviceMdibContainer(mdibbase.MdibContainer):
    """Device side implementation of an mdib.
     Do not modify containers directly, use transactions for that purpose.
     Transactions keep track of changes and initiate sending of update notifications to clients."""

    def __init__(self, sdc_definitions, log_prefix=None, waveform_source=None):
        """
        :param sdc_definitions: defaults to sdc11073.definitions_sdc.SDC_v1_Definitions
        :param log_prefix: a string
        :param waveform_source: an instance of an object that implements devicewaveform.AbstractWaveformSource
        """
        if sdc_definitions is None:
            sdc_definitions = SDC_v1_Definitions
        super(DeviceMdibContainer, self).__init__(sdc_definitions)
        self._logger = loghelper.getLoggerAdapter('sdc.device.mdib', log_prefix)
        self._sdcDevice = None
        self._trLock = Lock()  # transaction lock

        self.sequenceId = uuid.uuid4().urn  # this uuid identifies this mdib instance

        self._currentLocation = None  # or a SdcLocation instance
        self._annotators = {}
        self._current_transaction = None

        self.preCommitHandler = None  # preCommitHandler can modify transaction if needed before it is committed
        self.postCommitHandler = None  # postCommitHandler can modify mdib if needed after it is committed
        self._waveform_source = waveform_source or DefaultWaveformSource()

    @contextmanager
    def mdibUpdateTransaction(self, setDeterminationTime=True):
        # pylint: disable=protected-access
        with self._trLock:
            try:
                self._current_transaction = _MdibUpdateTransaction(self)
                yield self._current_transaction
                if callable(self.preCommitHandler):
                    self.preCommitHandler(self, self._current_transaction)  # pylint: disable=not-callable
                if self._current_transaction._error:
                    self._logger.info('mdibUpdateTransaction: transaction without updates!')
                else:
                    self._process_transaction(setDeterminationTime)
                    if callable(self.postCommitHandler):
                        self.postCommitHandler(self, self._current_transaction)  # pylint: disable=not-callable
            finally:
                self._current_transaction = None

    @contextmanager
    def _rt_sample_transaction(self):
        with self._trLock:
            with self.mdibLock:
                try:
                    self._current_transaction = _RtDataMdibUpdateTransaction(self)
                    yield self._current_transaction
                    if callable(self.preCommitHandler):
                        self.preCommitHandler(self, self._current_transaction)  # pylint: disable=not-callable
                    if self._current_transaction._error:
                        self._logger.info('_rtsampleTransaction: transaction without updates!')
                    else:
                        self._process_internal_rt_transaction()
                        if callable(self.postCommitHandler):
                            self.postCommitHandler(self, self._current_transaction)  # pylint: disable=not-callable
                finally:
                    self._current_transaction = None

    def _process_transaction(self, setDeterminationTime):
        mgr = self._current_transaction
        now = time.time()
        increment_mdib_version = False

        descr_updated = []
        descr_created = []
        descr_deleted = []
        descr_updated_states = []
        metric_updates = []
        alert_updates = []
        comp_updates = []
        ctxt_updates = []
        op_updates = []
        rt_updates = []

        # BICEPS: The version number is incremented by one every time the descriptive part changes
        if len(mgr.descriptorUpdates) > 0:
            self.mdDescriptionVersion += 1
            increment_mdib_version = True

        # BICEPS: The version number is incremented by one every time the state part changes.
        if len(mgr.metricStateUpdates) > 0 or len(mgr.alertStateUpdates) > 0 \
                or len(mgr.componentStateUpdates) > 0 or len(mgr.contextStateUpdates) > 0 \
                or len(mgr.operationalStateUpdates) > 0 or len(mgr.rtSampleStateUpdates) > 0:
            self.mdStateVersion += 1
            increment_mdib_version = True

        if increment_mdib_version:
            self.mdibVersion += 1

        # handle descriptors
        if len(mgr.descriptorUpdates) > 0:
            # need to know all to be deleted and to be created descriptors
            to_be_deleted = [old for old, new in mgr.descriptorUpdates.values() if new is None]
            to_be_created = [new for old, new in mgr.descriptorUpdates.values() if old is None]
            to_be_deleted_handles = [d.handle for d in to_be_deleted]
            to_be_created_handles = [d.handle for d in to_be_created]
            with self.mdibLock:

                def _updateCorrespondingState(descriptorContainer):
                    # add state to updated_states list and to corresponding notifications input
                    # => the state is always sent twice, a) in the description modification report and b)
                    # in the specific state update notification.
                    if descriptorContainer.isAlertDescriptor:
                        update_dict = mgr.alertStateUpdates
                    elif descriptorContainer.isComponentDescriptor:
                        update_dict = mgr.componentStateUpdates
                    elif descriptorContainer.isContextDescriptor:
                        update_dict = mgr.contextStateUpdates
                    elif descriptorContainer.isRealtimeSampleArrayMetricDescriptor:
                        update_dict = mgr.rtSampleStateUpdates
                    elif descriptorContainer.isMetricDescriptor:
                        update_dict = mgr.metricStateUpdates
                    elif descriptorContainer.isOperationalDescriptor:
                        update_dict = mgr.operationalStateUpdates
                    else:
                        raise RuntimeError(f'do not know how to handle {descriptorContainer.__class__.__name__}')
                    if descriptorContainer.isContextDescriptor:
                        update_dict = mgr.contextStateUpdates
                        all_states = self.contextStates.descriptorHandle.get(descriptorContainer.handle, [])
                        for st in all_states:
                            key = (descriptorContainer.Handle, st.Handle)
                            # check if state is already present in this transaction
                            state_update = update_dict.get(key)
                            if state_update is not None:
                                # the state has also been updated directly in transaction.
                                # update descriptor version
                                old_state, new_state = state_update
                            else:
                                old_state = st
                                new_state = old_state.mkCopy()
                                update_dict[key] = _TrItem(old_state, new_state)
                            new_state.descriptorContainer = descriptorContainer
                            new_state.incrementState()
                            new_state.updateDescriptorVersion()
                            descr_updated_states.append(new_state)
                    else:
                        # check if state is already present in this transaction
                        state_update = update_dict.get(descriptorContainer.handle)
                        new_state = None
                        if state_update is not None:
                            # the state has also been updated directly in transaction.
                            # update descriptor version
                            old_state, new_state = state_update
                            if new_state is None:
                                raise ValueError(f'state deleted? that should not be possible! handle = '
                                                 f'{descriptorContainer.handle}')
                            new_state.descriptorContainer = descriptorContainer
                            new_state.updateDescriptorVersion()
                        else:
                            old_state = self.states.descriptorHandle.getOne(descriptorContainer.handle, allowNone=True)
                            if old_state is not None:
                                new_state = old_state.mkCopy()
                                new_state.descriptorContainer = descriptorContainer
                                new_state.incrementState()
                                new_state.updateDescriptorVersion()
                                update_dict[descriptorContainer.handle] = _TrItem(old_state, new_state)
                        if new_state is not None:
                            descr_updated_states.append(new_state)

                def _incrementParentDescriptorVersion(descriptorContainer):
                    parentDescriptorContainer = self.descriptions.handle.getOne(descriptorContainer.parentHandle)
                    parentDescriptorContainer.incrementDescriptorVersion()
                    descr_updated.append(parentDescriptorContainer)
                    _updateCorrespondingState(parentDescriptorContainer)

                # handling only updated states here: If a descriptor is created, I assume that the application also creates the state in an transaction.
                # The state will then be transported via that notification report.
                # Maybe this needs to be reworked, but at the time of this writing it seems fine.
                for tr_item in mgr.descriptorUpdates.values():
                    origDescriptor, newDescriptor = tr_item.old, tr_item.new
                    if newDescriptor is not None:
                        newDescriptor.updateNode(setXsiType=True)
                        # DescriptionModificationReport also contains the states that are related to the descriptors.
                        # => if there is one, update its DescriptorVersion and add it to list of states that shall be sent
                        # (Assuming that context descriptors (patient, location) are never changed,
                        #  additional check for states in self.contextStates is not needed.
                        #  If this assumption is wrong, that functionality must be added!)
                        _updateCorrespondingState(newDescriptor)
                    if origDescriptor is None:
                        # this is a create operation
                        self._logger.debug('mdibUpdateTransaction: new descriptor Handle={}, DescriptorVersion={}',
                                           newDescriptor.handle, newDescriptor.DescriptorVersion)
                        descr_created.append(newDescriptor.mkCopy())
                        self.descriptions.addObjectNoLock(newDescriptor)
                        # R0033: A SERVICE PROVIDER SHALL increment pm:AbstractDescriptor/@DescriptorVersion by one if a direct child descriptor is added or deleted.
                        if newDescriptor.parentHandle is not None and \
                                newDescriptor.parentHandle not in to_be_created_handles:
                            # only update parent if it is not also created in this transaction
                            _incrementParentDescriptorVersion(newDescriptor)
                    elif newDescriptor is None:
                        # this is a delete operation
                        self._logger.debug('mdibUpdateTransaction: rm descriptor Handle={}, DescriptorVersion={}',
                                           origDescriptor.handle, origDescriptor.DescriptorVersion)
                        all_descriptors = self.getAllDescriptorsInSubTree(origDescriptor)
                        self._rmDescriptorsAndStates(all_descriptors)
                        descr_deleted.extend(all_descriptors)
                        # R0033: A SERVICE PROVIDER SHALL increment pm:AbstractDescriptor/@DescriptorVersion by one if a direct child descriptor is added or deleted.
                        if origDescriptor.parentHandle is not None and \
                                origDescriptor.parentHandle not in to_be_deleted_handles:
                            # only update parent if it is not also deleted in this transaction
                            _incrementParentDescriptorVersion(origDescriptor)
                    else:
                        # this is an update operation
                        descr_updated.append(newDescriptor)
                        self._logger.debug('mdibUpdateTransaction: update descriptor Handle={}, DescriptorVersion={}',
                                           newDescriptor.handle, newDescriptor.DescriptorVersion)
                        self.descriptions.replaceObjectNoLock(newDescriptor)

        # handle metric states
        if len(mgr.metricStateUpdates) > 0:
            with self.mdibLock:
                # self.mdibVersion += 1
                self._logger.debug('mdibUpdateTransaction: mdib version={}, metric updates = {}',
                                   self.mdibVersion,
                                   mgr.metricStateUpdates)
                for value in mgr.metricStateUpdates.values():
                    oldstate, newstate = value.old, value.new
                    try:
                        if setDeterminationTime and newstate.metricValue is not None:
                            newstate.metricValue.DeterminationTime = now
                        # replace the old container with the new one
                        self.states.removeObjectNoLock(oldstate)
                        self.states.addObjectNoLock(newstate)
                        metric_updates.append(newstate)
                    except RuntimeError:
                        self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                        raise

        # handle alert states
        if len(mgr.alertStateUpdates) > 0:
            with self.mdibLock:
                self._logger.debug('mdibUpdateTransaction: alert State updates = {}', mgr.alertStateUpdates)
                for value in mgr.alertStateUpdates.values():
                    oldstate, newstate = value.old, value.new
                    try:
                        if setDeterminationTime and newstate.isAlertCondition:
                            newstate.DeterminationTime = time.time()
                        newstate.updateNode()
                        # replace the old container with the new one
                        self.states.removeObjectNoLock(oldstate)
                        self.states.addObjectNoLock(newstate)
                        alert_updates.append(newstate)
                    except RuntimeError:
                        self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                        raise

            # handle component state states
        if len(mgr.componentStateUpdates) > 0:
            with self.mdibLock:
                # self.mdibVersion += 1
                self._logger.debug('mdibUpdateTransaction: component State updates = {}', mgr.componentStateUpdates)
                for value in mgr.componentStateUpdates.values():
                    oldstate, newstate = value.old, value.new
                    try:
                        newstate.updateNode()
                        # replace the old container with the new one
                        self.states.removeObjectNoLock(oldstate)
                        self.states.addObjectNoLock(newstate)
                        comp_updates.append(newstate)
                    except RuntimeError:
                        self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                        raise

        # handle context states
        if len(mgr.contextStateUpdates) > 0:
            with self.mdibLock:
                self._logger.debug('mdibUpdateTransaction: contextState updates = {}', mgr.contextStateUpdates)
                for value in mgr.contextStateUpdates.values():
                    oldstate, newstate = value.old, value.new
                    try:
                        ctxt_updates.append(newstate)
                        # replace the old container with the new one
                        self.contextStates.removeObjectNoLock(oldstate)
                        self.contextStates.addObjectNoLock(newstate)
                        newstate.updateNode()
                    except RuntimeError:
                        self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                        raise

        # handle operational states
        if len(mgr.operationalStateUpdates) > 0:
            with self.mdibLock:
                self._logger.debug('mdibUpdateTransaction: operationalState updates = {}', mgr.operationalStateUpdates)
                for value in mgr.operationalStateUpdates.values():
                    oldstate, newstate = value.old, value.new
                    try:
                        newstate.updateNode()
                        self.states.removeObjectNoLock(oldstate)
                        self.states.addObjectNoLock(newstate)
                        op_updates.append(newstate)
                    except RuntimeError:
                        self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                        raise

        # handle real time samples
        # important note: this transaction does not pull values from registered waveform providers!
        # Application is responsible for providing data.
        if len(mgr.rtSampleStateUpdates) > 0:
            with self.mdibLock:
                self._logger.debug('mdibUpdateTransaction: rtSample updates = {}', mgr.rtSampleStateUpdates)
                for value in mgr.rtSampleStateUpdates.values():
                    oldstate, newstate = value.old, value.new
                    try:
                        newstate.updateNode()
                        self.states.removeObjectNoLock(oldstate)
                        self.states.addObjectNoLock(newstate)
                        rt_updates.append(newstate)
                    except RuntimeError:
                        self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                        raise
                    except:
                        raise

        mdib_version_grp = self.mdib_version_group
        if self._sdcDevice is not None:
            if len(mgr.descriptorUpdates) > 0:
                updated = [d.mkCopy() for d in descr_updated]
                created = [d.mkCopy() for d in descr_created]
                deleted = [d.mkCopy() for d in descr_deleted]
                updated_states = [s.mkCopy() for s in descr_updated_states]
                self._sdcDevice.sendDescriptorUpdates(mdib_version_grp, updated=updated, created=created, deleted=deleted,
                                                      updated_states=updated_states)
            if len(metric_updates) > 0:
                updates = [s.mkCopy() for s in metric_updates]
                self._sdcDevice.sendMetricStateUpdates(mdib_version_grp, updates)
            if len(alert_updates) > 0:
                updates = [s.mkCopy() for s in alert_updates]
                self._sdcDevice.sendAlertStateUpdates(mdib_version_grp, updates)
            if len(comp_updates) > 0:
                updates = [s.mkCopy() for s in comp_updates]
                self._sdcDevice.sendComponentStateUpdates(mdib_version_grp, updates)
            if len(ctxt_updates) > 0:
                updates = [s.mkCopy() for s in ctxt_updates]
                self._sdcDevice.sendContextStateUpdates(mdib_version_grp, updates)
            if len(op_updates) > 0:
                updates = [s.mkCopy() for s in op_updates]
                self._sdcDevice.sendOperationalStateUpdates(mdib_version_grp, updates)
            if len(rt_updates) > 0:
                updates = [s.mkCopy() for s in rt_updates]
                self._sdcDevice.sendRealtimeSamplesStateUpdates(mdib_version_grp, updates)
        mgr.mdib_version = self.mdibVersion

    def _process_internal_rt_transaction(self):
        mgr = self._current_transaction
        # handle real time samples
        if len(mgr.rtSampleStateUpdates) > 0:
            self.mdibVersion += 1
            updates = []
            self._logger.debug('mdibUpdateTransaction: rtSample updates = {}', mgr.rtSampleStateUpdates)
            for value in mgr.rtSampleStateUpdates.values():
                oldstate, newstate = value.old, value.new
                try:
                    newstate.updateNode()
                    updates.append(newstate)
                except RuntimeError:
                    self._logger.warn('mdibUpdateTransaction: {} did not exist before!! really??', newstate)
                    raise
            # makes copies of all states for sending, so that they can't be affected by transactions after this one
            updates = [s.mkCopy() for s in updates]
            if self._sdcDevice is not None:
                self._sdcDevice.sendRealtimeSamplesStateUpdates(self.mdib_version_group, updates)

        mgr.mdib_version = self.mdibVersion

    def setSdcDevice(self, sdcDevice):
        self._sdcDevice = sdcDevice

    def setLocation(self, sdcLocation, validators=None):
        """
        This method updates only the mdib internal data!
        use the SdcDevice.setLocation method if you want to publish the address an the network.
        :param sdcLocation: a pysdc.location.SdcLocation instance
        :param validators: a list of pysdc.pmtypes.InstanceIdentifier objects or None
        """
        allLocationContexts = self.contextStates.NODETYPE.get(domTag('LocationContextState'), [])
        with self.mdibUpdateTransaction() as mgr:
            # set all to currently associated Locations to Disassociated
            associatedLocations = [l for l in allLocationContexts if
                                   l.ContextAssociation == pmtypes.ContextAssociation.ASSOCIATED]
            for l in associatedLocations:
                locationContext = mgr.getContextState(l.descriptorHandle, l.Handle)
                locationContext.ContextAssociation = pmtypes.ContextAssociation.DISASSOCIATED
                # UnbindingMdibVersion is the first version in which it is no longer bound ( == this version + 1)
                locationContext.UnbindingMdibVersion = self.mdibVersion + 1
                locationContext.BindingEndTime = time.time()

            descriptorContainer = self.descriptions.NODETYPE.getOne(domTag('LocationContextDescriptor'))

            self._currentLocation = mgr.getContextState(descriptorContainer.handle)  # this creates a new location state
            self._currentLocation.updateFromSdcLocation(sdcLocation)
            if validators is not None:
                self._currentLocation.Validator = validators

    def _createDescriptorContainer(self, cls, nodeName, handle, parentHandle, codedValue, safetyClassification,
                                   ext_extension=None):
        obj = cls(nsmapper=self.nsmapper,
                  nodeName=nodeName,
                  handle=handle,
                  parentHandle=parentHandle,
                  )
        obj.SafetyClassification = safetyClassification
        obj.Type = codedValue
        if ext_extension:
            obj.ext_Extension = etree_.Element(extTag('Extension'))
            for node in ext_extension:
                obj.ext_Extension.append(node)
        obj.updateNode()
        return obj

    def createVmdDescriptorContainer(self, handle, parentHandle, codedValue, safetyClassification, ext_extension=None):
        """
        This method creates an VmdDescriptorContainer with the given properties.
        If it is called within an transaction, the created object is added to transaction and clients will be notified.
        Otherwise the object is only added to mdib without sending notifications to clients!
        :param handle: Handle of the new container
        :param parentHandle: Handle of the parent
        :param codedValue: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safetyClassification: a pmtypes.SafetyClassification value
        :param ext_extension: list of etree.Element elements to be added to the Extension element
        :return: the created object
        """
        qNameTag = domTag('Vmd')
        cls = self.getDescriptorContainerClass(qNameTag)
        obj = self._createDescriptorContainer(cls, qNameTag, handle, parentHandle, codedValue, safetyClassification,
                                              ext_extension)
        if self._current_transaction is not None:
            self._current_transaction.createDescriptor(obj)
        else:
            self.descriptions.addObject(obj)
        return obj

    def createChannelDescriptorContainer(self, handle, parentHandle, codedValue, safetyClassification,
                                         ext_extension=None):
        """
        This method creates a ChannelDescriptorContainer with the given properties and optionally adds it to the mdib.
        If it is called within an transaction, the created object is added to transaction and clients will be notified.
        Otherwise the object is only added to mdib without sending notifications to clients!
        :param handle: Handle of the new container
        :param parentHandle: Handle of the parent
        :param codedValue: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safetyClassification: a pmtypes.SafetyClassification value
        :param ext_extension: list of etree.Element elements to be added to the Extension element
        :return: the created object
        """
        qNameTag = domTag('Channel')
        cls = self.getDescriptorContainerClass(qNameTag)
        obj = self._createDescriptorContainer(cls, qNameTag, handle, parentHandle, codedValue, safetyClassification,
                                              ext_extension)
        if self._current_transaction is not None:
            self._current_transaction.createDescriptor(obj)
        else:
            self.descriptions.addObject(obj)
        return obj

    def createStringMetricDescriptorContainer(self, handle, parentHandle, codedValue, safetyClassification, unit,
                                              metricAvailability=pmtypes.MetricAvailability.INTERMITTENT,
                                              metricCategory=pmtypes.MetricCategory.UNSPECIFIED, ext_extension=None):
        """
        This method creates a StringMetricDescriptorContainer with the given properties and optionally adds it to the mdib.
        If it is called within an transaction, the created object is added to transaction and clients will be notified.
        Otherwise the object is only added to mdib without sending notifications to clients!
        :param handle: Handle of the new container
        :param parentHandle: Handle of the parent
        :param codedValue: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safetyClassification: a pmtypes.SafetyClassification value
        :param unit: a CodedValue
        :param metricAvailability: pmtypes.MetricAvailability
        :param metricCategory: pmtypes.MetricCategory
        :param ext_extension: list of etree.Element elements to be added to the Extension element
        :return: the created object
        """
        qNameTag = domTag('Metric')
        qNameType = domTag('StringMetricDescriptor')
        cls = self.getDescriptorContainerClass(qNameType)
        obj = self._createDescriptorContainer(cls, qNameTag, handle, parentHandle, codedValue, safetyClassification,
                                              ext_extension)
        obj.Unit = unit
        obj.MetricAvailability = metricAvailability
        obj.MetricCategory = metricCategory
        obj.updateNode()
        if self._current_transaction is not None:
            self._current_transaction.createDescriptor(obj)
        else:
            self.descriptions.addObject(obj)
        return obj

    def createEnumStringMetricDescriptorContainer(self, handle, parentHandle, codedValue, safetyClassification,
                                                  unit, allowedValues,
                                                  metricAvailability=pmtypes.MetricAvailability.INTERMITTENT,
                                                  metricCategory=pmtypes.MetricCategory.UNSPECIFIED,
                                                  ext_extension=None):
        """
        This method creates an EnumStringMetricDescriptorContainer with the given properties and optionally adds it
        to the mdib.
        If it is called within an transaction, the created object is added to transaction and clients will be notified.
        Otherwise the object is only added to mdib without sending notifications to clients!
        :param handle: Handle of the new container
        :param parentHandle: Handle of the parent
        :param codedValue: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safetyClassification: a pmtypes.SafetyClassification value
        :param unit: pmtypes.CodedValue
        :param allowedValues:
        :param metricAvailability: pmtypes.MetricAvailability
        :param metricCategory: pmtypes.MetricCategory
        :param ext_extension: list of etree.Element elements to be added to the Extension element
        :return: the created object
        """
        qNameTag = domTag('Metric')
        qNameType = domTag('EnumStringMetricDescriptor')
        cls = self.getDescriptorContainerClass(qNameType)
        obj = self._createDescriptorContainer(cls, qNameTag, handle, parentHandle, codedValue, safetyClassification,
                                              ext_extension)
        obj.Unit = unit
        obj.MetricAvailability = metricAvailability
        obj.MetricCategory = metricCategory
        obj.AllowedValue = allowedValues
        obj.updateNode()
        if self._current_transaction is not None:
            self._current_transaction.createDescriptor(obj)
        else:
            self.descriptions.addObject(obj)
        return obj

    def createClockDescriptorContainer(self, handle, parentHandle, codedValue, safetyClassification,
                                       ext_extension=None):
        """
        This method creates a ClockDescriptorContainer with the given properties.
        If it is called within an transaction, the created object is added to transaction and clients will be notified.
        Otherwise the object is only added to mdib without sending notifications to clients!
        :param handle: Handle of the new container
        :param parentHandle: Handle of the parent
        :param codedValue: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safetyClassification: a pmtypes.SafetyClassification value
        :param ext_extension: list of etree.Element elements to be added to the Extension element
        :return: the created object
        """
        cls = self.getDescriptorContainerClass(domTag('ClockDescriptor'))
        obj = self._createDescriptorContainer(cls, domTag('Clock'), handle, parentHandle, codedValue,
                                              safetyClassification, ext_extension)
        if self._current_transaction is not None:
            self._current_transaction.createDescriptor(obj)
        else:
            self.descriptions.addObject(obj)
        return obj

    def addState(self, stateContainer, adjustStateVersion=True):
        if self._current_transaction is not None:
            self._current_transaction.addState(stateContainer, adjustStateVersion)
        else:
            if stateContainer.isContextState:
                if stateContainer.Handle in self.contextStates.handle:
                    raise ValueError('context state Handle {} already in mdib!'.format(stateContainer.Handle))
                if adjustStateVersion:
                    self.contextStates.setVersion(stateContainer)
            else:
                if stateContainer.descriptorHandle in self.states.descriptorHandle:
                    raise ValueError(
                        'state descriptorHandle {} already in mdib!'.format(stateContainer.descriptorHandle))
                if adjustStateVersion:
                    self.states.setVersion(stateContainer)

    def addMdsNode(self, mdsNode):
        """
        This method creates DescriptorContainers and StateContainers from the provided dom tree.
        If it is called within an transaction, the created objects are added to transaction and clients will be notified.
        Otherwise the objects are only added to mdib without sending notifications to clients!
        :param mdsNode: a node representing data of a complete mds
        :return: None
        """
        msg_reader = msgreader.MessageReader(self)
        descriptorContainers = msg_reader.readMdDescription(mdsNode)
        if self._current_transaction is not None:
            for descr in descriptorContainers:
                self._current_transaction.createDescriptor(descr)
        else:
            for descr in descriptorContainers:
                self.descriptions.addObject(descr)

        stateContainers = msg_reader.readMdState(mdsNode, additionalDescriptorContainers=descriptorContainers)
        for s in stateContainers:
            self.addState(s)
        self.mkStateContainersforAllDescriptors()

    # real time data handling
    def registerWaveformGenerator(self, descriptorHandle, wfGenerator):
        self._waveform_source.register_waveform_generator(self, descriptorHandle, wfGenerator)

    def setWaveformGeneratorActivationState(self, descriptorHandle, componentActivation):
        self._waveform_source.set_activation_state(self, descriptorHandle, componentActivation)

    def registerAnnotationGenerator(self, annotator, triggerHandle, annotatedHandles):
        self._waveform_source.register_annotation_generator(annotator, triggerHandle, annotatedHandles)

    def update_all_rt_samples(self):
        with self._rt_sample_transaction() as tr:
            self._waveform_source.update_all_realtime_samples(tr)

    def mkStateContainersforAllDescriptors(self):
        """The model requires that there is a state for every descriptor (exception: multi-states)
        Call this method to create missing states
        :return:
        """
        new_states = []
        for descr in self.descriptions.objects:
            if descr.handle not in self.states.descriptorHandle and descr.handle not in self.contextStates.descriptorHandle:
                state_cls = self.getStateClsForDescriptor(descr)
                if state_cls.isMultiState:
                    pass  # nothing to do, it is allowed to have no state
                else:
                    st = state_cls(self.nsmapper, descr)
                    # add some initial values where needed
                    if st.isAlertCondition:
                        st.DeterminationTime = time.time()
                    elif st.NODETYPE == domTag('AlertSystemState'):
                        st.LastSelfCheck = time.time()
                        st.SelfCheckCount = 1
                    elif st.NODETYPE == domTag('ClockState'):
                        st.LastSet = time.time()
                    st.updateNode()
                    if self._current_transaction is not None:
                        self._current_transaction.addState(st)
                    else:
                        self.states.addObject(st)
                        new_states.append(st)
        if new_states:
            self._updateStateObservables(new_states)

    @classmethod
    def fromMdibFile(cls, path, createLocationContextDescr=True, createPatientContextDescr=True,
                     protocol_definition=None, log_prefix=None):
        """
        An alternative constructor for the class
        :param path: the input file path for creating the mdib
        :param createLocationContextDescr: same as in fromString method
        :param createPatientContextDescr: same as in fromString method
        :param protocol_definition: an optional object derived from BaseDefinitions, forces usage of this definition
        :param log_prefix: a string or None
        :return: instance
        """
        with open(path, 'rb') as f:
            xml_text = f.read()
        return DeviceMdibContainer.fromString(xml_text, createLocationContextDescr, createPatientContextDescr,
                                              protocol_definition, log_prefix)

    @classmethod
    def fromString(cls, xml_text, createLocationContextDescr=True, createPatientContextDescr=True,
                   protocol_definition=None, log_prefix=None):
        """
        An alternative constructor for the class
        :param xml_text: the input string for creating the mdib
        :param createLocationContextDescr: if True, and the mdib does not contain a LocationContextDescriptor, it adds one
        :param createPatientContextDescr: if True, and the mdib does not contain a PatientContextDescriptor, it adds one
        :param protocol_definition: an optional object derived from BaseDefinitions, forces usage of this definition
        :param log_prefix: a string or None
        :return: instance
        """
        # get protocol definition that matches xml_text
        if protocol_definition is None:
            for definition_cls in ProtocolsRegistry.protocols:
                if definition_cls.ParticipantModelNamespace is not None and definition_cls.ParticipantModelNamespace.encode(
                        'utf-8') in xml_text:
                    protocol_definition = definition_cls
                    break
        if protocol_definition is None:
            raise ValueError('cannot create instance, no known BICEPS schema version identified')

        mdib = cls(protocol_definition, log_prefix=log_prefix)
        root = msgreader.MessageReader.getMdibRootNode(mdib.sdc_definitions, xml_text)
        mdib.sdc_definitions.xml_validator.assertValid(root)
        mdib.nsmapper.useDocPrefixes(root.nsmap)
        msg_reader = msgreader.MessageReader(mdib)
        # first make descriptions and add them to mdib, and then make states (they need already existing descriptions)
        descriptorContainers = msg_reader.readMdDescription(root)
        mdib.addDescriptionContainers(descriptorContainers)
        stateContainers = msg_reader.readMdState(root)
        mdib.addStateContainers(stateContainers)

        if createLocationContextDescr or createPatientContextDescr:
            # make sure we have exactly one PatientContext and one LocationContext Descriptor, depending of flags
            systemContextContainer = mdib.descriptions.NODETYPE.getOne(domTag('SystemContextDescriptor'))
            children = mdib.descriptions.parentHandle.get(systemContextContainer.handle)
            childdrenNodeNames = [ch.NODETYPE for ch in children]
            if createLocationContextDescr:
                qn = domTag('LocationContextDescriptor')
                if qn not in childdrenNodeNames:
                    mdib._logger.info('creating a LocationContextDescriptor')
                    descr_cls = mdib.getDescriptorContainerClass(qn)
                    lc = descr_cls(mdib.nsmapper, nodeName=domTag('LocationContext'),
                                   handle=uuid.uuid4().hex, parentHandle=systemContextContainer.handle)
                    lc.SafetyClassification = pmtypes.SafetyClassification.INF
                    mdib.descriptions.addObject(lc)
            if createPatientContextDescr:
                qn = domTag('PatientContextDescriptor')
                if qn not in childdrenNodeNames:
                    mdib._logger.info('creating a PatientContextDescriptor')
                    descr_cls = mdib.getDescriptorContainerClass(qn)
                    pc = descr_cls(mdib.nsmapper, nodeName=domTag('PatientContext'),
                                   handle=uuid.uuid4().hex, parentHandle=systemContextContainer.handle)
                    pc.SafetyClassification = pmtypes.SafetyClassification.INF
                    mdib.descriptions.addObject(pc)
        mdib.mkStateContainersforAllDescriptors()
        return mdib



import math
import itertools


def sinus(min_value, max_value, samples):
    delta = 2*math.pi/samples
    _values = [math.sin(i*delta) for i in range(samples)]  # -1 ... +1
    values = [(n+1)/2.0*(max_value-min_value) + min_value for n in _values]  # min ... max
    return values


def sawtooth(min_value, max_value, samples):
    delta = (max_value - min_value)/float(samples)
    values = [min_value + i*delta for i in range(samples)]
    return values


def triangle(min_value, max_value, samples):
    min_value = float(min_value)
    max_value = float(max_value)
    delta = (max_value - min_value)/float(samples) *2
    samples_cnt = int(samples/2)
    values = [min_value + i*delta for i in range(samples_cnt)]+ [max_value - i*delta for i in range(samples_cnt)]
    return values


    
class _WaveformGeneratorBase(object):
    def __init__(self, values_generator, min_value, max_value, waveformperiod, sampleperiod):
        if sampleperiod >= waveformperiod:
            raise ValueError('please choose a waveformperiod >> sampleperiod. currently use have wp={}, sp={}'.format(waveformperiod, sampleperiod))
        self.sampleperiod = sampleperiod
        samples = int(waveformperiod / sampleperiod)
        values = values_generator(min_value, max_value, samples)     
        wf_with_startflag = [ (v, i==0) for i, v in  enumerate(values)]
        self._values = wf_with_startflag
        self._generator = itertools.cycle(self._values)


    def nextSamples(self, count):
        return [next(self._generator) for i in range(count)]

    

class TriangleGenerator(_WaveformGeneratorBase):
    def __init__(self, min_value, max_value, waveformperiod, sampleperiod):
        super(TriangleGenerator, self).__init__(triangle, min_value, max_value, waveformperiod, sampleperiod)



class SawtoothGenerator(_WaveformGeneratorBase):
    def __init__(self,  min_value, max_value, waveformperiod, sampleperiod):
        super(SawtoothGenerator, self).__init__(sawtooth,  min_value, max_value, waveformperiod, sampleperiod)



class SinusGenerator(_WaveformGeneratorBase):
    def __init__(self,  min_value, max_value, waveformperiod, sampleperiod):
        super(SinusGenerator, self).__init__(sinus, min_value, max_value, waveformperiod, sampleperiod)



""" This module defines an ObservableProperty class.
An ObservablePropery must be declared as class attribute, similar to standard python properties.
You can bind callables to an ObservableProperty. The callable is called when the property value is set.

Example:
>>> class MyBaseClass(object):
>>>     prop1 = ObservableProperty(21)
>>>     prop2 = ObservableProperty(22)
>>>     
>>>     def __init__(self):
>>>         do_something()
>>> 
>>> class Observer(object):
>>>     def onProp1Changed(value):
>>>         print 'prop1=', value
>>>     def onProp21Changed(value):
>>>         print 'prop2=', value
>>> 
>>> actor = MyBaseClass()
>>> observer = Observer()
>>> bind(actor, prop1=observer.onProp1Changed, prop2=observer.onProp2Changed)

>>> actor.prop1=42
< prop1= 42
>>> actor.prop2='Hello World'
< prop2= Hello World

"""
import weakref
import inspect
from contextlib import contextmanager


class WeakRef:
    """ This Weaf Ref implementation allows to hold references to bound methods.
    => see http://stackoverflow.com/questions/599430/why-doesnt-the-weakref-work-on-this-bound-method"""
    def __init__ (self, item):
        self.reference = None
        self.method = None
        self.instance = None        
        try:
            self.method   = item.__func__.__name__
            self.instance = weakref.ref (item.__self__)
        except AttributeError:
            self.reference = weakref.ref (item)


    def getRef(self):
        if self.reference is not None:
            return self.reference ()
        instance = self.instance ()
        if instance == None:
            return None
        return getattr (instance, self.method)


    def __eq__(self, other):
        try:
            if self.reference is not None:
                return self.reference == other.reference
            return self.method == other.method and self.instance == other.instance
        except AttributeError:
            # other is of an unknown class 
            return False



class _ObservableValue(object):
    """ Implements the basic mechanism for an observable value. """
    def __init__(self, value, fireOnlyOnChangedValue=True):
        self.value = value
        self._fireOnlyOnChangedValue = fireOnlyOnChangedValue
        self._observers = []


    def setValue(self, value):
        if value == self.value and self._fireOnlyOnChangedValue:
            return
        self.value = value
        obsoleteRefs=[]
        # now call all listeners. Keep track of obsolete weak refeences
        for ref in self._observers[:]: # make a copy of list, content might change during iteration
            try:
                func = ref.getRef()
            except AttributeError: # no Weakref instance => strong reference, use ref directly
                func = ref
            if func is None:
                obsoleteRefs.append(ref)
            else:
                self._callFunc(func)
        for ref in obsoleteRefs:
            try:
                self._observers.remove(ref)
            except ValueError: # e.g. has been deleted by someone else in different thread.
                pass


    def _callFunc(self, func):
        func(self.value)


    def bind(self, func):
        self._observers.append(WeakRef(func))


    def strongbind(self, func):
        self._observers.append(func)


    def unbind(self, func):
        funcRef = WeakRef(func)
        for ref in self._observers:
            if ref == func or ref == funcRef:
                self._observers.remove(ref)
                break


    def unbindAll(self):
        del self._observers[:]



class ObservableProperty(object):
    """ stores data in parent obj """
    def __init__(self, defaultValue=None, fireOnlyOnChangedValue=True):
        self._defaultValue = defaultValue
        self._fireOnlyOnChangedValue = fireOnlyOnChangedValue


    def _getInstanceData(self, obj):
        # see if we already have a _PropertyInstanceData dictionary injected in obj
        # otherwise inject it
        try:
            lookup =  obj._Property2InstanceData
        except AttributeError:
            obj._Property2InstanceData = dict()
            lookup = obj._Property2InstanceData
        
        # see if we already have a data instance for my property instance and class instance
        # otherwise create one
        try:    
            return lookup[self]
        except KeyError:
            lookup[self] = _ObservableValue(self._defaultValue, self._fireOnlyOnChangedValue)
            return lookup[self]


    def __get__(self,  obj, objtype):
        return self if obj is None else self._getInstanceData(obj).value


    def __set__(self, obj, value):
        if obj is None: 
            self._defaultvalue = value
        else:
            self._getInstanceData(obj).setValue(value)


    def __delete__(self, obj):
        pass


    def bind(self, obj, func):
        self._getInstanceData(obj).bind(func)


    def strongbind(self, obj, func):
        self._getInstanceData(obj).strongbind(func)


    def unbind(self, obj, func):
        self._getInstanceData(obj).unbind(func)


    def unbindAll(self, obj):
        self._getInstanceData(obj).unbindAll()


    def __repr__(self):
        return 'ObservableProperty at 0x{1:X}, default value= {0}'.format(self._defaultValue, id(self))


def _findProperty(obj, name):
    """ Helper that looks in class hierarchy for matching member
    """
    classes = inspect.getmro(obj.__class__) # getmro returns a tuple of class base classes, including class, in method resolution order
    
    for cls in classes:   # find the first class that has the expected member
        try:
            return cls.__dict__[name]
        except KeyError:
            pass
    raise KeyError(name) # if no class matches, raise KeyError
    

def bind (obj, **kwargs):
    """ bind callables with a weak reference.
    Use this bind method for all 'normal' callables like functions or methods.
    The advantage is that the garbage collector can remove objects even if they are referenced by ObservableProperty.
    ObservableProperty silently removes the callable if it no longer exists.
    This method does not work with lambda expressions! 
    :param obj: an object with ObservableProperty member(s)
    :param **kwargs: name of parameter must match the name of an ObservableProperty, value must be a callable."""
    for name, func in kwargs.items():
        p = _findProperty(obj, name)
        p.bind(obj, func)


def strongbind (obj, **kwargs):
    """ bind callables with a strong reference.
     This method also works with lambda expressions, but you must unbind the callable before the garbage collector can delete it."""
    for name, func in kwargs.items():
        p = _findProperty(obj, name)
        p.strongbind(obj, func)


def unbind (obj, **kwargs):
    """ unbind callables that were bound before.
    :param obj: an object with ObservableProperty member(s)
    :param **kwargs: name of parameter must match the name of an ObservableProperty, value must be a callable.
                    Unbinding an unknown callable is allowed, in this cases nothing changes. """
    for name, func in kwargs.items():
        p = _findProperty(obj, name)
        p.unbind(obj, func)


def unbindAll (obj, *propertyNames):
    """ unbind all callables that were bound before.
    :param obj: an object with ObservableProperty member(s)
    :param *propertyNames: list of strings , each string names an ObservableProperty.
    """
    for name in propertyNames:
        p = _findProperty(obj, name)
        p.unbindAll(obj)


@contextmanager
def boundContext(obj, **kwargs):
    """ context manager for bind / unbind sequence."""
    bind(obj, **kwargs)
    try:
        yield
    finally:
        unbind(obj, **kwargs)


@contextmanager
def strongboundContext(obj, **kwargs):
    """ context manager for strongbind / unbind sequence."""
    strongbind(obj, **kwargs)
    try:
        yield
    finally:
        unbind(obj, **kwargs)




import dataclasses
import os
import ssl


@dataclasses.dataclass
class SSLContextContainer:
    client_context: ssl.SSLContext
    server_context: ssl.SSLContext


def mk_ssl_contexts_from_folder(ca_folder,
                                private_key='userkey.pem',
                                certificate='usercert.pem',
                                ca_public_key='cacert.pem',
                                cyphers_file=None,
                                ssl_passwd=None) -> SSLContextContainer:
    """Convenience method for easy creation of SSL context, assuming all needed files are in the same folder.
    Create an ssl context from files 'userkey.pem', 'usercert.pem', and optional 'cacert.pem' and cyphers file
    :param ca_folder: base path of all files
    :param private_key: name of the private key file of the user
    :param certificate: name of the signed certificate of the user
    :param ca_public_key: name of public key of the certificate authority that signed the certificate; if given,
                          verify_mode of ssl contexts in the return value will be set to CERT_REQUIRED
    :param cyphers_file: optional file that contains a cyphers string; comments are possible, start line with '#'
    :param ssl_passwd: optional password string
    :return: container of SSLContext instances i.e. client_ssl_context and server_ssl_context
    """
    certfile = os.path.join(ca_folder, certificate)
    keyfile = os.path.join(ca_folder, private_key)
    if ca_public_key:
        cafile = os.path.join(ca_folder, ca_public_key)
    else:
        cafile = None
    if cyphers_file:
        with open(os.path.join(ca_folder, cyphers_file)) as f:
            while True:
                # allow comment lines, starting with #
                cyphers = f.readline()
                if len(cyphers) == 0:  # end of file reached without having found a valid line
                    cyphers = None
                    break
                cyphers = cyphers.strip()
                cyphers = cyphers.rstrip('\n')
                cyphers = cyphers.rstrip('\r')
                if len(cyphers) > 0 and not cyphers.startswith("#"):
                    break
    else:
        cyphers = None
    return mk_ssl_contexts(keyfile, certfile, cafile, cyphers, ssl_passwd)


def mk_ssl_contexts(key_file,
                    cert_file,
                    ca_file,
                    cyphers=None,
                    ssl_passwd=None) -> SSLContextContainer:
    """Convenience method for easy creation of SSL context.
    Create an ssl context from files 'userkey.pem', 'usercert.pem', 'cacert.pem' and optional 'cyphers.json'
    :param key_file: the private key pem file of the user
    :param cert_file: the signed certificate of the user
    :param ca_file: optional public key of the certificate authority that signed the certificate; if given,
                    verify_mode of ssl contexts in the return value will be set to CERT_REQUIRED
    :param cyphers: optional cyphers string
    :param ssl_passwd: optional password string
    :return: container of SSLContext instances i.e. client_ssl_context and server_ssl_context
    """
    client_ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    server_ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    client_ssl_context.check_hostname = False
    client_ssl_context.load_cert_chain(certfile=cert_file, keyfile=key_file, password=ssl_passwd)
    if cyphers is not None:
        client_ssl_context.set_ciphers(cyphers)
    if ca_file:
        client_ssl_context.verify_mode = ssl.CERT_REQUIRED
        client_ssl_context.load_verify_locations(ca_file)

    server_ssl_context.load_cert_chain(certfile=cert_file, keyfile=key_file, password=ssl_passwd)
    if cyphers is not None:
        server_ssl_context.set_ciphers(cyphers)
    if ca_file:
        server_ssl_context.verify_mode = ssl.CERT_REQUIRED
        server_ssl_context.load_verify_locations(ca_file)

    return SSLContextContainer(client_context=client_ssl_context, server_context=server_ssl_context)



from collections import defaultdict
from lxml import etree as etree_
from .. import pysoap
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import msgTag, dpwsTag, nsmap

from .sdcservicesimpl import DPWSPortTypeImpl
from .sdcservicesimpl import WSDLMessageDescription, WSDLOperationBinding
from .sdcservicesimpl import _wsdl_ns, _mkWsdlTwowayOperation

_msg = Prefix.MSG.prefix

def _tw2i(textwidth_string):
    """ text width to int"""
    lookup = {'xs': 0, 's':1, 'm':2, 'l':3, 'xl':4, 'xxl':5, None: 999}
    return lookup[textwidth_string]

def _calcNoL(text):
    # definition of a line in Participant Model:
    # ...a line is defined as the content of the text from either the beginning of the text or the beginning of
    # a previous line until the next occurance of period mark, question mark, exclamation mark, or paragraph.
    # TBD: naive approach?
    return len(text.split('\n'))


def _textWidth_filter(localized_texts, width):
    i_width = _tw2i(width)
    candidates = [v for v in localized_texts if _tw2i(v.TextWidth) <= i_width]
    if candidates:
        candidates.sort(key=lambda obj:  _tw2i(obj.TextWidth) or -1)
    return candidates

def _n_o_l_filter(localized_texts, n_o_l):
    candidates = [v for v in localized_texts if v.n_o_l <= n_o_l]
    if candidates:
        candidates.sort(key=lambda obj: obj.n_o_l or -1)
    return candidates


class LocalizationStorage():
    def __init__(self, localizedTexts=None):
        self._localizedTexts = defaultdict(list) # key = handle, value = list of pmtypes.LocalizedText objects
        if localizedTexts:
            self.add(*localizedTexts)

    def add(self, *localizedTexts):
        for t in localizedTexts:
            self._localizedTexts[t.Ref].append(t)

    def filterLocalizedTexts(self, requestedHandles, requestedVersion, requestedLangs, textWidths, numberOfLines):
        """

        :param requestedHandles: list of handles
        :param requestedVersion: an integer or None
        :param requestedLangs: list of language strings
        :param textWidths: a list of integers, 0...n
        :param numberOfLines: a list of integers, 0...n
        :return: a list of LocalizedText instances
        """
        # make integers for textWidths and numberOfLines
        if textWidths is None:
            textWidths = []
        if numberOfLines is None:
            numberOfLines = []
        if requestedHandles is None:
            requestedHandles = []
        i_nls = [int(l) for l in numberOfLines]

        if len(requestedHandles) == 0:
            # If there is no Ref ELEMENT given in the request MESSAGE, then all texts are returned in
            # msg:GetLocalizedTextResponse/msg:Text
            handles = list(self._localizedTexts.keys())
        else:
            # If there is at least one Ref ELEMENT given, then msg:GetLocalizedTextResponse/msg:Text contains all texts
            # that match the Ref ELEMENTs of the msg:GetLocalizedText request MESSAGE.
            handles = requestedHandles

        # create a flat list of all localized texts with the requested handles
        texts = []
        for h in handles:
            try:
                texts.extend(self._localizedTexts[h])
            except KeyError:
                pass

        # filter languages:
        if requestedLangs is not None and len(requestedLangs) > 0:
            texts = [t for t in texts if t.Lang in requestedLangs]

        # filter requested versions. We need to do it per language, therefore create a lookup with (ref,lang) as key
        tmp_dict = defaultdict(list)
        for t in texts:
            tmp_dict[(t.Ref, t.Lang)].append(t)
        texts = []

        if requestedVersion is None:
            # determine the highest available Version in the storage
            all = []
            for v in self._localizedTexts.values():
                all.extend(v)
            all = [a.Version for a in all if a.Version is not None]
            requestedVersion = max(all)
        # If the referenced text is not available in the specific version, then
        # msg:GetLocalizedTextResponse/msg:Text is empty
        for key, value_list in tmp_dict.items():
            texts.extend([v for v in value_list if v.Version == requestedVersion])

        # - If there is no NumberOfLines ELEMENT given in the request MESSAGE, then all texts independent of the number
        #   of lines are returned in msg:GetLocalizedTextResponse/msg:Text.
        # - If there is at least one NumberOfLines ELEMENT given, msg:GetLocalizedTextResponse/msg:Text contains texts
        #   that match the number of lines defined by the NumberOfLines ELEMENTs of the msg:GetLocalizedText request
        #   MESSAGE. Matching in this case means that the number of lines in the text is less or equal to the
        #   NumberOfLines ELEMENTs.

        if len(textWidths) > 0 or len(numberOfLines) > 0:
            if len(numberOfLines) > 0:
                # calculate number of lines for every localized text and add is as member to the object
                for t in texts:
                    t.n_o_l = _calcNoL(t.text)

            # create again dictionary by ref and language:
            tmp_dict = defaultdict(list)
            for t in texts:
                tmp_dict[(t.Ref, t.Lang)].append(t)
            tmp = []

            if len(textWidths) > 0 and len(numberOfLines) > 0:
                # now find for each combination of (width, lines) list the best match
                for key, value_list in tmp_dict.items():
                    candidates = []
                    for w in textWidths:
                        candidates1 = _textWidth_filter(value_list, w) # returns sorted list of smaller elements
                        for l in i_nls:
                            candidates2 = _n_o_l_filter(candidates1, l)
                            if len(candidates2) > 0:
                                candidates2.sort(key=lambda obj: obj.TextWidth*obj.n_o_l) # sort by area size
                                tmp.append(candidates2[-1])  # use largest one
            elif len(textWidths) > 0:
                # filter only text widths
                for key, value_list in tmp_dict.items():
                    for w in textWidths:
                        candidates = _textWidth_filter(value_list, w)  # returns sorted list of smaller elements
                        if candidates:
                            tmp.append(candidates[-1])  # use largest one

            elif len(numberOfLines) > 0:
                # filter only number of lines
                for key, value_list in tmp_dict.items():
                    for l in i_nls:
                        candidates = _n_o_l_filter(value_list, l)
                        if candidates:
                            tmp.append(candidates[-1])  # use largest one
            texts = list(tmp)
        return texts


    def getSupportedlanguages(self):
        texts = self._flatList()
        result = set()
        for text in texts:
            result.add(str(text.Lang))
        return list(result)

    def _flatList(self, refList=None):
        if refList is None:
            # If there is no Ref ELEMENT given in the request MESSAGE, then all texts are returned in
            # msg:GetLocalizedTextResponse/msg:Text
            handles = list(self._localizedTexts.keys())
        else:
            # If there is at least one Ref ELEMENT given, then msg:GetLocalizedTextResponse/msg:Text contains all texts
            # that match the Ref ELEMENTs of the msg:GetLocalizedText request MESSAGE.
            handles = refList

        # create a flat list of all localized texts with the requested handles
        texts = []
        for h in handles:
            try:
                texts.extend(self._localizedTexts[h])
            except KeyError:
                pass
        return texts


class LocalizationService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('GetLocalizedText', ('{}:GetLocalizedText'.format(_msg),)),
                               WSDLMessageDescription('GetLocalizedTextResponse', ('{}:GetLocalizedTextResponse'.format(_msg),)),
                               WSDLMessageDescription('GetSupportedLanguages', ('{}:GetSupportedLanguages'.format(_msg),)),
                               WSDLMessageDescription('GetSupportedLanguagesResponse', ('{}:GetSupportedLanguagesResponse'.format(_msg),)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('GetLocalizedText', 'literal', 'literal'),
                             WSDLOperationBinding('GetSupportedLanguages', 'literal', 'literal'),)

    def __init__(self, port_type_string, sdcDevice):
        super(LocalizationService, self).__init__(port_type_string, sdcDevice )
        self.register_soapActionCallback(self._mdib.sdc_definitions.Actions.GetLocalizedText, self._onGetLocalizedText)
        self.register_soapActionCallback(self._mdib.sdc_definitions.Actions.GetSupportedLanguages, self._onGetSupportedLanguages)
        self.localizationStorage = LocalizationStorage()

    def _onGetLocalizedText(self, httpHeader, request):  # pylint:disable=unused-argument
        self._logger.debug('_onGetLocalizedText')
        requestedHandles = request.bodyNode.xpath('*/msg:Ref/text()', namespaces=nsmap) # handle strings 0...n
        requestedVersions = request.bodyNode.xpath('*/msg:Version/text()', namespaces=nsmap) # unsigned long int 0..1
        requestedLangs = request.bodyNode.xpath('*/msg:Lang/text()', namespaces=nsmap) # unsigned long int 0..n
        textWidths = request.bodyNode.xpath('*/msg:TextWidth/text()', namespaces=nsmap) # strings 0..n
        numberOfLines = request.bodyNode.xpath('*/msg:NumberOfLines/text()', namespaces=nsmap) # int 0..n
        # make integers for textWidths and numberOfLines
        i_tws = [_tw2i(w) for w in textWidths]
        i_nls = [int(l) for l in numberOfLines]

        if len(requestedVersions) > 0:
            # If the referenced text is not available in the specific version, then
            # msg:GetLocalizedTextResponse/msg:Text is empty
            requestedVersion = int(requestedVersions[0])
        else:
            requestedVersion = None

        texts = self.localizationStorage.filterLocalizedTexts(requestedHandles,
                                                              requestedVersion,
                                                              requestedLangs,
                                                              i_tws, i_nls)

        # create the response
        nsmapper = self._mdib.nsmapper
        responseSoapEnvelope = pysoap.soapenvelope.Soap12Envelope(
            nsmapper.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG))
        replyAddress = request.address.mkReplyAddress(action=self._getActionString('GetLocalizedTextResponse'))
        responseSoapEnvelope.addHeaderObject(replyAddress)
        getLocalizedTextResponseNode = etree_.Element(msgTag('GetLocalizedTextResponse'))
        self._mdib.mdib_version_group.update_node(getLocalizedTextResponseNode)

        for text in texts:
            getLocalizedTextResponseNode.append(text.asEtreeNode(msgTag('Text'), nsmap=None))
        responseSoapEnvelope.addBodyElement((getLocalizedTextResponseNode))
        return responseSoapEnvelope


    def _onGetSupportedLanguages(self, httpHeader, request):  # pylint:disable=unused-argument
        self._logger.debug('_onGetSupportedLanguages')
        languages = self.localizationStorage.getSupportedlanguages()

        nsmapper = self._mdib.nsmapper
        responseSoapEnvelope = pysoap.soapenvelope.Soap12Envelope(
            nsmapper.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG))
        replyAddress = request.address.mkReplyAddress(action=self._getActionString('GetSupportedLanguagesResponse'))
        responseSoapEnvelope.addHeaderObject(replyAddress)
        getSupportedLanguagesResponseNode = etree_.Element(msgTag('GetSupportedLanguagesResponse'))
        self._mdib.mdib_version_group.update_node(getSupportedLanguagesResponseNode)

        for lang in languages:
            n = etree_.SubElement(getSupportedLanguagesResponseNode, msgTag('Lang'))
            n.text = lang
        responseSoapEnvelope.addBodyElement((getSupportedLanguagesResponseNode))
        return responseSoapEnvelope


    def addWsdlPortType(self, parentNode):
        """
        add wsdl:portType node to parentNode.
        xml looks like this:
        <wsdl:portType name="GetService" dpws:DiscoveryType="dt:ServiceProvider">
          <wsdl:operation name="GetMdState">
            <wsdl:input message="msg:GetLocalizedText"/>
            <wsdl:output message="msg:GetLocalizedTextResponse"/>
          </wsdl:operation>
          <wsp:Policy>
            <dpws:Profile wsp:Optional="true"/>
          </wsp:Policy>
          ...
        </wsdl:portType>
        :param parentNode:
        :return:
        """
        if 'dt' in parentNode.nsmap:
            portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                         attrib={'name': self.port_type_string,
                                                 dpwsTag('DiscoveryType'):'dt:ServiceProvider'})
        else:
            portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns, 'portType'),
                                         attrib={'name': self.port_type_string})
        _mkWsdlTwowayOperation(portType, operationName='GetLocalizedText')
        _mkWsdlTwowayOperation(portType, operationName='GetSupportedLanguages')



import time
from collections import namedtuple, OrderedDict
from io import BytesIO
from lxml import etree as etree_

from .. import pysoap
from ..namespaces import Prefix_Namespace as Prefix
from ..namespaces import msgTag, domTag, s12Tag, wsxTag, wseTag, dpwsTag, mdpwsTag, nsmap
from .. import pmtypes
from .. import loghelper
from .exceptions import InvalidActionError, FunctionNotImplementedError, InvalidMessageError
_msg_prefix = Prefix.MSG.prefix

_wsdl_ns = Prefix.WSDL.namespace
_wsdl_message = etree_.QName(_wsdl_ns, 'message')
_wsdl_part = etree_.QName(_wsdl_ns, 'part')
_wsdl_operation = etree_.QName(_wsdl_ns, 'operation')

_wsp_ns = 'http://www.w3.org/ns/ws-policy'
_wsp_prefix = 'wsp'

# DiscoveryType, only used in SDC
_dt = "http://standards.ieee.org/downloads/11073/11073-10207-2017"

_WSDL_S12 = "http://schemas.xmlsoap.org/wsdl/soap12/" # old soap 12 namespace, used in wsdl 1.1. used only for wsdl


# WSDL Generation:
# types to allow declaration of a wsdl data per service
WSDLMessageDescription  = namedtuple('WSDLMessageDescription', 'name parameters ')
WSDLOperationBinding = namedtuple('WSDLOperationBinding', 'name input output')


def etreeFromFile(path):
    parser = etree_.ETCompatXMLParser()
    with open(path, 'rb') as f:
        xml_text = f.read()
    return etree_.fromstring(xml_text, parser=parser, base_url=path)


class SOAPActionRegistry(object):
    def __init__(self, log_prefix=None):
        self._soapActionCallbacks = {}
        self._getCallbacks = {}
        self._logger = loghelper.getLoggerAdapter('sdc.device.{}'.format(self.__class__.__name__), log_prefix)

    def register_soapActionCallback(self, action, fn):
        self._soapActionCallbacks[action] = fn

    def register_getCallback(self, path, query, fn):
        if path.endswith('/'):
            path = path[:-1]
        self._getCallbacks[(path, query)] = fn

    def _getActionHandler(self, action):
        """ returns a callable or None"""
        return self._soapActionCallbacks.get(action)

    def getActions(self):
        """ returns a list of action strings that can be handled."""
        return  list(self._soapActionCallbacks.keys())


class SOAPActionDispatcher(SOAPActionRegistry):

    def dispatchSoapRequest(self, path, httpHeader, soapEnvelope):
        begin = time.monotonic()
        action = soapEnvelope.address.action
        fn = self._getActionHandler(action)
        if fn is None:
            raise InvalidActionError(soapEnvelope)
        returnedSoapEnvelope = fn(httpHeader, soapEnvelope)
        duration = time.monotonic() - begin
        self._logger.debug('incoming soap action "{}" to {}: duration={:.3f}sec.', action, path, duration)
        return returnedSoapEnvelope

    def dispatchGetRequest(self, parseResult, httpHeader): #pylint:disable=unused-argument
        begin = time.monotonic()
        path = parseResult.path
        if path.endswith('/'):
            path = path[:-1]
        key = (path, parseResult.query)
        fn = self._getCallbacks.get(key)
        if fn is not None:
            self._logger.debug('dispatchGetRequest:path="{}" ,function="{}"', key, fn.__name__)
            result = fn()
            duration = time.monotonic() - begin
            self._logger.debug('dispatchGetRequest:duration="{:.4f}"', duration)
            return result
        else:
            self._logger.error('dispatchGetRequest:path="{}" ,no handler found!', key)
            raise KeyError('dispatchGetRequest:path="{}" ,no handler found!'.format(key))


class _SOAPActionDispatcherWithSubDispatchers(SOAPActionDispatcher):
    """ receiver of all messages"""

    def __init__(self, subDispatchers=None):
        super(_SOAPActionDispatcherWithSubDispatchers, self).__init__()
        self.subDispatchers = subDispatchers or [] # chained SOAPActionDispatcher


    def _getActionHandler(self, action):
        """ returns a callable or None"""
        fn = self._soapActionCallbacks.get(action)
        if fn is not None:
            return fn
        for sd in self.subDispatchers:
            fn = sd._getActionHandler(action)
            if fn is not None:
                return fn
        return None


    def getActions(self):
        """ returns a list of action strings that can be handled."""
        actions =  list(self._soapActionCallbacks.keys())
        for sd in self.subDispatchers:
            actions.extend(sd.getActions())
        return actions

    def __repr__(self):
        return '{} actions={}'.format(self.__class__.__name__, self._soapActionCallbacks.keys())



class EventService(_SOAPActionDispatcherWithSubDispatchers):
    """ A service that offers subscriptions"""
    def __init__(self, sdcDevice, subDispatchers, offeredSubscriptions):
        super(EventService, self).__init__(subDispatchers)

        self._sdcDevice = sdcDevice
        self._subscriptionsManager = sdcDevice.subscriptionsManager
        self._offeredSubscriptions = offeredSubscriptions
        self.register_soapActionCallback('{}/Subscribe'.format(Prefix.WSE.namespace), self._onSubscribe)
        self.register_soapActionCallback('{}/Unsubscribe'.format(Prefix.WSE.namespace), self._onUnsubscribe)
        self.register_soapActionCallback('{}/GetStatus'.format(Prefix.WSE.namespace), self._onGetStatus)
        self.register_soapActionCallback('{}/Renew'.format(Prefix.WSE.namespace), self._onRenewStatus)
        self.epr = None

    def dispatchSoapRequest(self, path, httpHeader, soapEnvelope):
        if self._sdcDevice.shallValidate:
            soapEnvelope.validate_envelope(self._sdcDevice.xml_validator)
        response_envelope = super().dispatchSoapRequest(path, httpHeader, soapEnvelope)
        if self._sdcDevice.shallValidate:
            response_envelope.validate_envelope(self._sdcDevice.xml_validator)
        return response_envelope

    def _onSubscribe(self, httpHeader, soapEnvelope):
        subscriptionFilters = soapEnvelope.bodyNode.xpath(
            "//wse:Filter[@Dialect='{}/Action']".format(Prefix.DPWS.namespace),
            namespaces=nsmap)
        if len(subscriptionFilters) != 1:
            raise InvalidMessageError(request=soapEnvelope, detail="wse:Subscribe/wse:Filter not provided")
        else:
            sfilters = subscriptionFilters[0].text
            for sfilter in sfilters.split():
                if sfilter not in self._offeredSubscriptions:
                    raise Exception('{}::{}: "{}" is not in offered subscriptions: {}'.format(self.__class__.__name__,
                                                                                              self.epr,
                                                                                              sfilter,
                                                                                              self._offeredSubscriptions))
        returnedSoapEnvelope = self._subscriptionsManager.onSubscribeRequest(httpHeader, soapEnvelope, self.epr)
        return returnedSoapEnvelope

    def _onUnsubscribe(self, httpHeader, soapEnvelope): #pylint:disable=unused-argument
        returnedSoapEnvelope = self._subscriptionsManager.onUnsubscribeRequest(soapEnvelope)
        return returnedSoapEnvelope

    def _onGetStatus(self, httpHeader, soapEnvelope): #pylint:disable=unused-argument
        returnedSoapEnvelope = self._subscriptionsManager.onGetStatusRequest(soapEnvelope)
        return returnedSoapEnvelope

    def _onRenewStatus(self, httpHeader, soapEnvelope): #pylint:disable=unused-argument
        returnedSoapEnvelope = self._subscriptionsManager.onRenewRequest(soapEnvelope)
        return returnedSoapEnvelope


class DPWSHostedService(EventService):
    """ An Endpoint (url) with one or more DPWS Types"""

    def __init__(self, sdcDevice, base_urls, path_suffix, subDispatchers, offeredSubscriptions):
        """
        :param base_urls: urlparse.SplitResult instances. They define the base addresses of this service
        """
        super(DPWSHostedService, self).__init__(sdcDevice, subDispatchers, offeredSubscriptions)

        self._base_urls = base_urls
        self._mdib = sdcDevice.mdib
        self._my_port_types = [p.port_type_string for p in subDispatchers]
        self._wsdlString = self._mkWsdlString()
        my_uuid = base_urls[0].path
        for s in subDispatchers:
            s.hostingService = self
        self.epr = '/{}/{}'.format(my_uuid, path_suffix)  # end point reference
        endpointReferencesList = []
        for addr in base_urls:
            endpointReferencesList.append(pysoap.soapenvelope.WsaEndpointReferenceType('{}/{}'.format(addr.geturl(), path_suffix)))
        porttype_ns = sdcDevice.mdib.sdc_definitions.PortTypeNamespace
        # little bit ugly: normalizeXMLText needs bytes, not string. and it looks for namespace in "".
        porttype_ns = sdcDevice.mdib.sdc_definitions.normalizeXMLText(b'"'+porttype_ns.encode('utf-8')+b'"')[1:-1].decode('utf-8')
        self.hostedInf = pysoap.soapenvelope.DPWSHosted(endpointReferencesList=endpointReferencesList,
                                                        typesList=[
                                                            etree_.QName(porttype_ns, p) for p in self._my_port_types],
        serviceId=self._my_port_types[0])

        self.register_soapActionCallback('{}/GetMetadata/Request'.format(Prefix.WSX.namespace), self._onGetMetaData)
        self.register_getCallback(path=self.epr, query='wsdl', fn=self._onGetWSDL)


    def _onGetWSDL(self):
        """ return wsdl"""
        return self._wsdlString


    def _mkWsdlString(self):
        #biceps_schema = self._sdcDevice.mdib.bicepsSchema
        sdc_definitions = self._sdcDevice.mdib.sdc_definitions
        my_nsmap = Prefix.partialMap(Prefix.MSG, Prefix.PM, Prefix.WSA, Prefix.WSE, Prefix.DPWS)
        my_nsmap['tns'] = Prefix.SDC.namespace
        my_nsmap['dt'] = _dt
        porttype_prefix = 'tns'
        my_nsmap['wsdl'] = _wsdl_ns
        my_nsmap['s12'] = _WSDL_S12
        my_nsmap[_wsp_prefix] = _wsp_ns
        wsdl_definitions = etree_.Element(etree_.QName(_wsdl_ns, 'definitions'),
                                          nsmap=my_nsmap,
                                          attrib={'targetNamespace': sdc_definitions.PortTypeNamespace})

        types = etree_.SubElement(wsdl_definitions, etree_.QName(_wsdl_ns, 'types'))
        #remove annotations from schemas, this reduces wsdl size from 280kb to 100kb!
        extSchema_ = etreeFromFile(sdc_definitions.ExtensionPointSchemaFile)
        extSchema = self._removeAnnotations(extSchema_)
        pmSchema_ = etreeFromFile(sdc_definitions.ParticipantModelSchemaFile)
        pmSchema = self._removeAnnotations(pmSchema_)
        bmmSchema_ = etreeFromFile(sdc_definitions.MessageModelSchemaFile)
        bmmSchema = self._removeAnnotations(bmmSchema_)
        types.append(extSchema)
        types.append(pmSchema)
        types.append(bmmSchema)
        # append all message nodes
        for s in self.subDispatchers:
            s.addWsdlMessages(wsdl_definitions)
        for s in self.subDispatchers:
            s.addWsdlPortType(wsdl_definitions)
        for s in self.subDispatchers:
            s.addWsdlBinding(wsdl_definitions, porttype_prefix)

        s =  etree_.tostring(wsdl_definitions)
        return sdc_definitions.denormalizeXMLText(s)


    def _removeAnnotations(self, root_node):
        remove_annotations_string = b"""<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                                      xmlns:xs="http://www.w3.org/2001/XMLSchema">
          <xsl:output method="xml" indent="yes"/>

          <xsl:template match="@* | node()">
            <xsl:copy>
              <xsl:apply-templates select="@* | node()"/>
            </xsl:copy>
          </xsl:template>

          <xsl:template match="xs:annotation" />
        </xsl:stylesheet>"""
        remove_annotations_doc = etree_.parse(BytesIO(remove_annotations_string))
        remove_annotations_xslt = etree_.XSLT(remove_annotations_doc)
        return remove_annotations_xslt(root_node).getroot()


    def _onGetMetaData(self, httpHeader, request):
        _nsm = self._mdib.nsmapper
        response = pysoap.soapenvelope.Soap12Envelope(_nsm.docNssmap)
        replyAddress = request.address.mkReplyAddress('http://schemas.xmlsoap.org/ws/2004/09/mex/GetMetadata/Response')
        response.addHeaderObject(replyAddress)

        metaDataNode = etree_.Element(wsxTag('Metadata'), nsmap=_nsm.docNssmap)

        # Relationship
        metaDataSectionNode = etree_.SubElement(metaDataNode,
                                                wsxTag('MetadataSection'),
                                                attrib={'Dialect': '{}/Relationship'.format(Prefix.DPWS.namespace)})

        relationshipNode = etree_.SubElement(metaDataSectionNode,
                                             dpwsTag('Relationship'),
                                             attrib={'Type': '{}/host'.format(Prefix.DPWS.namespace)})
        self._sdcDevice.dpwsHost.asEtreeSubNode(relationshipNode)

        self.hostedInf.asEtreeSubNode(relationshipNode)

        metaDataSectionNode = etree_.SubElement(metaDataNode,
                                                wsxTag('MetadataSection'),
                                                attrib={'Dialect': _wsdl_ns})
        locationNode = etree_.SubElement(metaDataSectionNode,
                                         wsxTag('Location'))
        # determine the correct location of wsdl, depending on call
        host = httpHeader['Host'] # this is the address that was called.
        my_baseUrls = [u for u in self._base_urls if u.netloc == host]
        my_baseUrl = my_baseUrls[0] if len(my_baseUrls) > 0 else self._base_urls[0]
        locationNode.text = '{}://{}{}/?wsdl'.format(my_baseUrl.scheme,
                                               my_baseUrl.netloc,
                                               self.epr                                               )
        response.addBodyElement(metaDataNode)
        return response

    def __repr__(self):
        return '{} epr={} Porttypes={}'.format(self.__class__.__name__, self.epr, [dp.port_type_string for dp in self.subDispatchers])


class DPWSPortTypeImpl(SOAPActionRegistry):
    """ Base class of all PortType implementations"""
    WSDLOperationBindings = () # overwrite in derived classes
    WSDLMessageDescriptions = () # overwrite in derived classes

    def __init__(self, port_type_string, sdcDevice):
        """
        :param port_type_string: port type without namespace, e.g 'Get'
        :param sdcDevice:
        """
        super(DPWSPortTypeImpl, self).__init__()
        self.port_type_string = port_type_string
        self._sdcDevice = sdcDevice
        self._mdib = sdcDevice.mdib
        self.hostingService = None  # the parent

    def _getActionString(self, methodName):
        actions_lookup = self._mdib.sdc_definitions.Actions
        return getattr(actions_lookup, methodName)

    def addWsdlPortType(self, parentNode):
        raise NotImplementedError

    def __repr__(self):
        return '{} Porttype={} actions={}'.format(self.__class__.__name__, self.port_type_string,
                                                  self._soapActionCallbacks.keys())


    def addWsdlMessages(self, parentNode):
        """
        add wsdl:message node to parentNode.
        xml looks like this:
        <wsdl:message name="GetMdDescription">
            <wsdl:part element="msg:GetMdDescription" name="parameters" />
        </wsdl:message>
        :param parentNode:
        :return:
        """
        for msg in self.WSDLMessageDescriptions:
            elem = etree_.SubElement(parentNode, _wsdl_message, attrib={'name': msg.name})
            for elementName in msg.parameters:
                etree_.SubElement(elem, _wsdl_part,
                                  attrib={'name': 'parameters',
                                          'element': elementName})


    def addWsdlBinding(self, parentNode, porttype_prefix):
        """
        add wsdl:binding node to parentNode.
        xml looks like this:
        <wsdl:binding name="GetBinding" type="msg:Get">
            <s12:binding style="document" transport="http://schemas.xmlsoap.org/soap/http" />
            <wsdl:operation name="GetMdib">
                <s12:operation soapAction="http://p11073-10207/draft6/msg/2016/12/08/Get/GetMdib" />
                <wsdl:input>
                    <s12:body use="literal" />
                </wsdl:input>
                <wsdl:output>
                    <s12:body use="literal" />
                </wsdl:output>
            </wsdl:operation>
            ...
        </wsdl:binding>
        :param parentNode:
        :param porttype_prefix:
        :return:
        """
        v_ref = self._sdcDevice.mdib.sdc_definitions
        wsdl_binding = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'binding'),
                                     attrib={'name': self.port_type_string+'Binding', 'type': '{}:{}'.format(porttype_prefix, self.port_type_string)})
        s12_binding = etree_.SubElement(wsdl_binding, etree_.QName(_WSDL_S12, 'binding'),
                                        attrib={'style':'document', 'transport': 'http://schemas.xmlsoap.org/soap/http'})
        #ToDo: wsp:policy?
        for wsdl_op in self.WSDLOperationBindings:
            wsdl_operation = etree_.SubElement(wsdl_binding, etree_.QName(_wsdl_ns, 'operation'), attrib={'name': wsdl_op.name})
            s12_operation = etree_.SubElement(wsdl_operation, etree_.QName(_WSDL_S12, 'operation'),
                                              attrib={'soapAction': '{}/{}/{}'.format(v_ref.ActionsNamespace, self.port_type_string, wsdl_op.name)})
            if wsdl_op.input is not None:
                wsdl_input = etree_.SubElement(wsdl_operation, etree_.QName(_wsdl_ns, 'input'))
                etree_.SubElement(wsdl_input, etree_.QName(_WSDL_S12, 'body'), attrib={'use': wsdl_op.input})
            if wsdl_op.output is not None:
                wsdl_output = etree_.SubElement(wsdl_operation, etree_.QName(_wsdl_ns, 'output'))
                etree_.SubElement(wsdl_output, etree_.QName(_WSDL_S12, 'body'), attrib={'use': wsdl_op.output})
        _addPolicy_dpwsProfile(wsdl_binding)


def __mkWsdlOperation(parentNode, operationName, inputMessageName, outputMessageName,  fault):
    elem = etree_.SubElement(parentNode, _wsdl_operation, attrib={'name': operationName})
    if inputMessageName is not None:
        etree_.SubElement(elem,  etree_.QName(_wsdl_ns, 'input'),
                          attrib={'message': '{}:{}'.format('tns', inputMessageName),
                                 })
    if outputMessageName is not None:
        etree_.SubElement(elem,  etree_.QName(_wsdl_ns, 'output'),
                          attrib={'message': '{}:{}'.format('tns', outputMessageName),
                                 })
    if fault is not None:
        faultName, messageName, action = fault # unpack 3 parameters
        etree_.SubElement(elem,  etree_.QName(_wsdl_ns, 'fault'),
                          attrib={'name': faultName,
                                  'message': '{}:{}'.format('tns', messageName),
                                  })
    return elem


def _mkWsdlTwowayOperation(parentNode, operationName, inputMessageName=None, outputMessageName=None,  fault=None):
    # has input and output
    input_MessageName = inputMessageName or operationName # defaults to operation name
    output_MessageName = outputMessageName or operationName+'Response' # defaults to operation name + "Response"
    return __mkWsdlOperation(parentNode, operationName=operationName, inputMessageName=input_MessageName,
                            outputMessageName=output_MessageName, fault=fault)

def _mkWsdlOnewayOperation(parentNode, operationName, outputMessageName=None, fault=None):
    # has only output
    output_MessageName = outputMessageName or operationName # defaults to operation name
    return __mkWsdlOperation(parentNode, operationName=operationName, inputMessageName=None,
                            outputMessageName=output_MessageName, fault=fault)

def _addPolicy_dpwsProfile(parentNode):
    """
    :param parentNode:
    :return: <wsp:Policy>
            <dpws:Profile wsp:Optional="true"/>
            <mdpws:Profile wsp:Optional="true"/>
          </wsp:Policy>
    """
    wsp_policyNode = etree_.SubElement(parentNode, etree_.QName(_wsp_ns, 'Policy'), attrib=None)
    _ = etree_.SubElement(wsp_policyNode, dpwsTag('Profile'), attrib={etree_.QName(_wsp_ns, 'Optional'): 'true'})
    _ = etree_.SubElement(wsp_policyNode, mdpwsTag('Profile'), attrib={etree_.QName(_wsp_ns, 'Optional'): 'true'})


class GetService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('GetMdState', ('{}:GetMdState'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetMdStateResponse', ('{}:GetMdStateResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetMdib', ('{}:GetMdib'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetMdibResponse', ('{}:GetMdibResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetMdDescription', ('{}:GetMdDescription'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetMdDescriptionResponse', ('{}:GetMdDescriptionResponse'.format(_msg_prefix),)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('GetMdState', 'literal', 'literal'),
                             WSDLOperationBinding('GetMdib', 'literal', 'literal'),
                             WSDLOperationBinding('GetMdDescription', 'literal', 'literal'),)

    def __init__(self, port_type_string, sdcDevice):
        super(GetService, self).__init__(port_type_string, sdcDevice )
        actions = self._mdib.sdc_definitions.Actions
        self.register_soapActionCallback(actions.GetMdState, self._onGetMdState)
        self.register_soapActionCallback(actions.GetMdib, self._onGetMdib)
        self.register_soapActionCallback(actions.GetMdDescription, self._onGetMdDescription)


    def _onGetMdState(self, httpHeader, request):  # pylint:disable=unused-argument
        self._logger.debug('_onGetMdState')
        requestedHandles = request.bodyNode.xpath('*/msg:HandleRef/text()', namespaces=nsmap)
        if len(requestedHandles) > 0:
            self._logger.info('_onGetMdState requested Handles:{}', requestedHandles)

        # get the requested state containers from mdib
        stateContainers = []
        with self._mdib.mdibLock:
            if len(requestedHandles) == 0:
                # MessageModel: If the HANDLE reference list is empty, all states in the MDIB SHALL be included in the result list.
                for stateContainer in self._mdib.states.objects:
                    stateContainers.append(stateContainer)
                if self._sdcDevice.contextstates_in_getmdib:
                    for stateContainer in self._mdib.contextStates.objects:
                        stateContainers.append(stateContainer)
            else:
                if self._sdcDevice.contextstates_in_getmdib:
                    for handle in requestedHandles:
                        try:
                            # If a HANDLE reference does match a multi state HANDLE, the corresponding multi state SHALL be included in the result list
                            stateContainers.append(self._mdib.contextStates.handle.getOne(handle))
                        except RuntimeError:
                            # If a HANDLE reference does match a descriptor HANDLE, all states that belong to the corresponding descriptor SHALL be included in the result list
                            stateContainers.extend(self._mdib.states.descriptorHandle.get(handle, []))
                            stateContainers.extend(self._mdib.contextStates.descriptorHandle.get(handle, []))
                else:
                    for handle in requestedHandles:
                        stateContainers.extend(self._mdib.states.descriptorHandle.get(handle, []))

                self._logger.info('_onGetMdState requested Handles:{} found {} states', requestedHandles,
                                  len(stateContainers))

            # build response
            nsmapper = self._mdib.nsmapper
            responseSoapEnvelope = pysoap.soapenvelope.Soap12Envelope(
                nsmapper.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG))
            replyAddress = request.address.mkReplyAddress(action=self._getActionString('GetMdStateResponse'))
            responseSoapEnvelope.addHeaderObject(replyAddress)
            getMdStateResponseNode = etree_.Element(msgTag('GetMdStateResponse'), nsmap=nsmap)
            self._mdib.mdib_version_group.update_node(getMdStateResponseNode)

            mdStateNode = etree_.Element(msgTag('MdState'), attrib=None, nsmap=self._mdib.nsmapper.docNssmap)
            for stateContainer in stateContainers:
                mdStateNode.append(stateContainer.mkStateNode())

            getMdStateResponseNode.append(mdStateNode)
            responseSoapEnvelope.addBodyElement(getMdStateResponseNode)
        self._logger.debug('_onGetMdState returns {}', lambda: responseSoapEnvelope.as_xml(pretty=False))
        return responseSoapEnvelope

    def _onGetMdib(self, httpHeader, request):  # pylint:disable=unused-argument
        self._logger.debug('_onGetMdib')
        nsmapper = self._mdib.nsmapper
        responseSoapEnvelope = pysoap.soapenvelope.Soap12Envelope(
            nsmapper.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG))
        replyAddress = request.address.mkReplyAddress(action=self._getActionString('GetMdibResponse'))
        responseSoapEnvelope.addHeaderObject(replyAddress)
        if self._sdcDevice.contextstates_in_getmdib:
            mdibNode, mdib_version_group = self._mdib.reconstructMdibWithContextStates()
        else:
            mdibNode, mdib_version_group = self._mdib.reconstructMdib()
        getMdibResponseNode = etree_.Element(msgTag('GetMdibResponse'), nsmap=Prefix.partialMap(Prefix.MSG, Prefix.PM))
        mdib_version_group.update_node(getMdibResponseNode)
        getMdibResponseNode.append(mdibNode)
        responseSoapEnvelope.addBodyElement(getMdibResponseNode)
        self._logger.debug('_onGetMdib returns {}', lambda: responseSoapEnvelope.as_xml(pretty=False))
        return responseSoapEnvelope

    def _onGetMdDescription(self, httpHeader, request):  # pylint:disable=unused-argument
        """
        MdDescription comprises the requested set of MDS descriptors. Which MDS descriptors are included depends on the msg:GetMdDescription/msg:HandleRef list:
        - If the HANDLE reference list is empty, all MDS descriptors SHALL be included in the result list.
        - If a HANDLE reference does match an MDS descriptor, it SHALL be included in the result list.
        - If a HANDLE reference does not match an MDS descriptor (any other descriptor), the MDS descriptor that is in the parent tree of the HANDLE reference SHOULD be included in the result list.
        """
        # currently this implementation only supports a single mds.
        # => if at least one handle matches any descriptor, the one mds is returned, otherwise empty payload

        self._logger.debug('_onGetMdDescription')
        requestedHandles = request.bodyNode.xpath('*/msg:HandleRef/text()', namespaces=nsmap)
        if len(requestedHandles) > 0:
            self._logger.info('_onGetMdDescription requested Handles:{}', requestedHandles)
        includeMds = True if len(requestedHandles) == 0 else False  # if we have handles, we need to check them
        for h in requestedHandles:
            if self._sdcDevice.mdib.descriptions.handle.getOne(h, allowNone=True) is not None:
                includeMds = True
                break
        my_namespaces = self._sdcDevice.mdib.nsmapper.partialMap(Prefix.S12, Prefix.WSA, Prefix.MSG, Prefix.PM)
        responseSoapEnvelope = pysoap.soapenvelope.Soap12Envelope(my_namespaces)
        replyAddress = request.address.mkReplyAddress(action=self._getActionString('GetMdDescriptionResponse'))
        responseSoapEnvelope.addHeaderObject(replyAddress)

        getMdDescriptionResponseNode = etree_.Element(msgTag('GetMdDescriptionResponse'),
                                                      nsmap=nsmap)

        if includeMds:
            mdDescriptionNode, mdib_version_group = self._mdib.reconstructMdDescription()
            mdDescriptionNode.tag = msgTag('MdDescription')  # rename according to message
            mdib_version_group.update_node(getMdDescriptionResponseNode)
        else:
            mdDescriptionNode = etree_.Element(msgTag('MdDescription'))
            self._mdib.mdib_version_group.update_node(getMdDescriptionResponseNode)

        getMdDescriptionResponseNode.append(mdDescriptionNode)
        responseSoapEnvelope.addBodyElement(getMdDescriptionResponseNode)
        self._logger.debug('_onGetMdDescription returns {}', lambda: responseSoapEnvelope.as_xml(pretty=False))
        return responseSoapEnvelope


    def addWsdlPortType(self, parentNode):
        """
        add wsdl:portType node to parentNode.
        xml looks like this:
        <wsdl:portType name="GetService" dpws:DiscoveryType="dt:ServiceProvider">
          <wsdl:operation name="GetMdState">
            <wsdl:input message="msg:GetMdState"/>
            <wsdl:output message="msg:GetMdStateResponse"/>
          </wsdl:operation>
          <wsp:Policy>
            <dpws:Profile wsp:Optional="true"/>
          </wsp:Policy>
          ...
        </wsdl:portType>
        :param parentNode:
        :return:
        """
        if 'dt' in parentNode.nsmap:
            portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                         attrib={'name': self.port_type_string,
                                                 dpwsTag('DiscoveryType'):'dt:ServiceProvider'})
        else:
            portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns, 'portType'),
                                         attrib={'name': self.port_type_string})
        _mkWsdlTwowayOperation(portType, operationName='GetMdState')
        _mkWsdlTwowayOperation(portType, operationName='GetMdib')
        _mkWsdlTwowayOperation(portType, operationName='GetMdDescription')


class ContainmentTreeService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('GetDescriptor', ('{}:GetDescriptor'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetDescriptorResponse', ('{}:GetDescriptorResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetContainmentTree', ('{}:GetContainmentTreeResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetContainmentTreeResponse', ('{}:GetContainmentTreeResponse'.format(_msg_prefix),)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('GetDescriptor', 'literal', 'literal'),
                             WSDLOperationBinding('GetContainmentTree', 'literal', 'literal'))


    def __init__(self, port_type_string, sdcDevice):
        super(ContainmentTreeService, self).__init__(port_type_string, sdcDevice)
        actions = self._mdib.sdc_definitions.Actions
        self.register_soapActionCallback(actions.GetContainmentTree, self._onGetContainmentTree)
        self.register_soapActionCallback(actions.GetDescriptor, self._onGetDescriptor)


    def _onGetContainmentTree(self, httpHeader, request):
        #ToDo: implement, currently method only raises a soap fault
        raise FunctionNotImplementedError(request)


    def _onGetDescriptor(self, httpHeader, request):
        #ToDo: implement, currently method only raises a soap fault
        raise FunctionNotImplementedError(request)


    def addWsdlPortType(self, parentNode):
        portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                     attrib={'name':self.port_type_string,
                                             dpwsTag('DiscoveryType'):'dt:ServiceProvider'})
        _mkWsdlTwowayOperation(portType, operationName='GetDescriptor')
        _mkWsdlTwowayOperation(portType, operationName='GetContainmentTree')


class SetService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('Activate', ('{}:Activate'.format(_msg_prefix),)),
                               WSDLMessageDescription('ActivateResponse', ('{}:ActivateResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetString', ('{}:SetString'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetStringResponse', ('{}:SetStringResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetComponentState', ('{}:SetComponentState'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetComponentStateResponse', ('{}:SetComponentStateResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetAlertState', ('{}:SetAlertState'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetAlertStateResponse', ('{}:SetAlertStateResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetMetricState', ('{}:SetMetricState'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetMetricStateResponse', ('{}:SetMetricStateResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetValue', ('{}:SetValue'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetValueResponse', ('{}:SetValueResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('OperationInvokedReport', ('{}:OperationInvokedReport'.format(_msg_prefix),)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('Activate', 'literal', 'literal'), # fault?
                             WSDLOperationBinding('SetString', 'literal', 'literal'),  # fault?
                             WSDLOperationBinding('SetComponentState', 'literal', 'literal'), # fault?
                             WSDLOperationBinding('SetAlertState', 'literal', 'literal'), # fault?
                             WSDLOperationBinding('SetMetricState', 'literal', 'literal'), # fault?
                             WSDLOperationBinding('SetValue', 'literal', 'literal'), # fault?
                             WSDLOperationBinding('OperationInvokedReport', None, 'literal'),
                             )

    def __init__(self, port_type_string, sdcDevice):
        super(SetService, self).__init__(port_type_string, sdcDevice)
        actions = self._mdib.sdc_definitions.Actions
        self.register_soapActionCallback(actions.Activate, self._onActivate)
        self.register_soapActionCallback(actions.SetValue, self._onSetValue)
        self.register_soapActionCallback(actions.SetString, self._onSetString)
        self.register_soapActionCallback(actions.SetMetricState, self._onSetMetricState)
        self.register_soapActionCallback(actions.SetAlertState, self._onSetAlertState)
        self.register_soapActionCallback(actions.SetComponentState, self._onSetComponentState)


    def _onActivate(self, httpHeader, request): #pylint:disable=unused-argument
        """Handler for Active calls.
        It enques an operation and generates the expected operation invoked report. """
        return self._handleOperationRequest(request, 'ActivateResponse')

    def _onSetValue(self, httpHeader, request): #pylint:disable=unused-argument
        """Handler for SetValue calls.
        It enques an operation and generates the expected operation invoked report. """
        self._logger.info('_onSetValue')
        ret = self._handleOperationRequest(request, 'SetValueResponse')
        self._logger.info('_onSetValue done')
        return ret

    def _onSetString(self, httpHeader, request):  # pylint:disable=unused-argument
        """Handler for SetString calls.
        It enques an operation and generates the expected operation invoked report."""
        self._logger.debug('_onSetString')
        return self._handleOperationRequest(request, 'SetStringResponse')

    def _onSetMetricState(self, httpHeader, request):  # pylint:disable=unused-argument
        """Handler for SetMetricState calls.
        It enques an operation and generates the expected operation invoked report."""
        self._logger.debug('_onSetMetricState')
        return self._handleOperationRequest(request, 'SetMetricStateResponse')


    def _onSetAlertState(self, httpHeader, request):  # pylint:disable=unused-argument
        """Handler for SetMetricState calls.
        It enques an operation and generates the expected operation invoked report."""
        self._logger.debug('_onSetAlertState')
        return self._handleOperationRequest(request, 'SetAlertStateResponse')


    def _onSetComponentState(self, httpHeader, request):  # pylint:disable=unused-argument
        """Handler for SetMetricState calls.
        It enques an operation and generates the expected operation invoked report."""
        self._logger.debug('_onSetAlertState')
        return self._handleOperationRequest(request, 'SetComponentStateResponse')


    def _handleOperationRequest(self, request, responseName):  # pylint:disable=unused-argument
        """
        It enques an operation and generate the expected operation invoked report.
        :param request:
        :param responseName:
        :return:
        """
        response = pysoap.soapenvelope.Soap12Envelope(self._mdib.nsmapper.partialMap(Prefix.S12, Prefix.WSA))
        replyAddress = request.address.mkReplyAddress(action=self._getActionString(responseName))
        response.addHeaderObject(replyAddress)
        replyBodyNode = etree_.Element(msgTag(responseName),
                                       nsmap=Prefix.partialMap(Prefix.MSG))
        self._mdib.mdib_version_group.update_node(replyBodyNode)

        invocationInfoNode = etree_.SubElement(replyBodyNode, msgTag('InvocationInfo'))

        transactionIdNode = etree_.SubElement(invocationInfoNode, msgTag('TransactionId'))
        invocationStateNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationState'))

        errorTexts = []

        operationHandleRefs = request.bodyNode.xpath('*/msg:OperationHandleRef/text()', namespaces=nsmap)
        operation = None
        if len(operationHandleRefs) == 1:
            operationHandleRef = operationHandleRefs[0]
            operation = self._sdcDevice.getOperationByHandle(operationHandleRef)
            if operation is None:
                errorTexts.append('operation not known: "{}"'.format(operationHandleRef))
        else:
            errorTexts.append('no OperationHandleRef found in Request')

        if errorTexts:
            self._logger.warn('_handleOperationRequest: errorTexts = {}'.format(errorTexts))

            invocationStateNode.text = pmtypes.InvocationState.FAILED
            transactionIdNode.text = '0'
            operationErrorNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationError'))
            operationErrorNode.text = pmtypes.InvocationError.INVALID_VALUE
            operationErrorMessageNode = etree_.SubElement(invocationInfoNode,
                                                          msgTag('InvocationErrorMessage'))
            operationErrorMessageNode.text = '; '.join(errorTexts)
        else:
            self._logger.info('_handleOperationRequest: enqueued')
            transactionId = self._sdcDevice.enqueueOperation(operation, request)
            transactionIdNode.text = str(transactionId)
            invocationStateNode.text = pmtypes.InvocationState.WAIT

        response.addBodyElement(replyBodyNode)
        return response

    def addWsdlPortType(self, parentNode):
        portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                     attrib={'name':self.port_type_string,
                                             dpwsTag('DiscoveryType'):'dt:ServiceProvider',
                                             wseTag('EventSource'): 'true'})
        _mkWsdlTwowayOperation(portType, operationName='Activate')
        _mkWsdlTwowayOperation(portType, operationName='SetString')
        _mkWsdlTwowayOperation(portType, operationName='SetComponentState')
        _mkWsdlTwowayOperation(portType, operationName='SetAlertState')
        _mkWsdlTwowayOperation(portType, operationName='SetMetricState')
        _mkWsdlTwowayOperation(portType, operationName='SetValue')
        _mkWsdlOnewayOperation(portType, operationName='OperationInvokedReport')


class WaveformService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('Waveform', ('{}:WaveformStreamReport'.format(_msg_prefix),)),)
    WSDLOperationBindings = (WSDLOperationBinding('Waveform', None, 'literal'),)

    def addWsdlPortType(self, parentNode):
        portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                     attrib={'name':self.port_type_string,
                                             dpwsTag('DiscoveryType'):'dt:ServiceProvider',
                                             wseTag('EventSource'): 'true'})
        _mkWsdlOnewayOperation(portType, operationName='Waveform')


class StateEventService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('EpisodicAlertReport', ('{}:EpisodicAlertReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('SystemErrorReport', ('{}:SystemErrorReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('PeriodicAlertReport', ('{}:PeriodicAlertReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('EpisodicComponentReport', ('{}:EpisodicComponentReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('PeriodicOperationalStateReport', ('{}:PeriodicOperationalStateReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('PeriodicComponentReport', ('{}:PeriodicComponentReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('EpisodicOperationalStateReport', ('{}:EpisodicOperationalStateReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('PeriodicMetricReport', ('{}:PeriodicMetricReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('EpisodicMetricReport', ('{}:EpisodicMetricReport'.format(_msg_prefix),)),
                               )

    WSDLOperationBindings = (WSDLOperationBinding('EpisodicAlertReport', None, 'literal'),
                             WSDLOperationBinding('SystemErrorReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicAlertReport', None, 'literal'),
                             WSDLOperationBinding('EpisodicComponentReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicOperationalStateReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicComponentReport', None, 'literal'),
                             WSDLOperationBinding('EpisodicOperationalStateReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicMetricReport', None, 'literal'),
                             WSDLOperationBinding('EpisodicMetricReport', None, 'literal'),
                             )

    def addWsdlPortType(self, parentNode):
        portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                     attrib={'name':self.port_type_string,
                                             dpwsTag('DiscoveryType'):'dt:ServiceProvider',
                                             wseTag('EventSource'): 'true'})
        _mkWsdlOnewayOperation(portType, operationName='EpisodicAlertReport')
        _mkWsdlOnewayOperation(portType, operationName='SystemErrorReport')
        _mkWsdlOnewayOperation(portType, operationName='PeriodicAlertReport')
        _mkWsdlOnewayOperation(portType, operationName='EpisodicComponentReport')
        _mkWsdlOnewayOperation(portType, operationName='PeriodicOperationalStateReport')
        _mkWsdlOnewayOperation(portType, operationName='PeriodicComponentReport')
        _mkWsdlOnewayOperation(portType, operationName='EpisodicOperationalStateReport')
        _mkWsdlOnewayOperation(portType, operationName='PeriodicMetricReport')
        _mkWsdlOnewayOperation(portType, operationName='EpisodicMetricReport')


class ContextService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('SetContextState', ('{}:SetContextState'.format(_msg_prefix),)),
                               WSDLMessageDescription('SetContextStateResponse', ('{}:SetContextStateResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetContextStates', ('{}:GetContextStates'.format(_msg_prefix),)),
                               WSDLMessageDescription('GetContextStatesResponse', ('{}:GetContextStatesResponse'.format(_msg_prefix),)),
                               WSDLMessageDescription('EpisodicContextReport', ('{}:EpisodicContextReport'.format(_msg_prefix),)),
                               WSDLMessageDescription('PeriodicContextReport', ('{}:PeriodicContextReport'.format(_msg_prefix),)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('SetContextState', 'literal', 'literal'), #ToDo: generate wsdl:fault
                             WSDLOperationBinding('GetContextStates', 'literal', 'literal'),
                             WSDLOperationBinding('EpisodicContextReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicContextReport', None, 'literal'),
                             )


    def __init__(self, port_type_string, sdcDevice):
        super(ContextService, self).__init__(port_type_string, sdcDevice)
        actions = self._mdib.sdc_definitions.Actions
        self.register_soapActionCallback(actions.SetContextState, self._onSetContextState)
        self.register_soapActionCallback(actions.GetContextStates, self._onGetContextStates)

    def _onSetContextState(self, httpHeader, request):  # pylint:disable=unused-argument
        """ enqueues an operation and returns a 'wait' reponse."""
        response = pysoap.soapenvelope.Soap12Envelope(
            self._mdib.nsmapper.partialMap(Prefix.S12, Prefix.PM, Prefix.WSA, Prefix.MSG))
        replyAddress = request.address.mkReplyAddress(action=self._getActionString('SetContextStateResponse'))
        response.addHeaderObject(replyAddress)
        replyBodyNode = etree_.Element(msgTag('SetContextStateResponse'),
                                       nsmap=Prefix.partialMap(Prefix.MSG),
                                       attrib={'SequenceId': self._mdib.sequenceId,
                                               'MdibVersion': str(self._mdib.mdibVersion)})
        invocationInfoNode = etree_.SubElement(replyBodyNode,
                                               msgTag('InvocationInfo'))
        transactionIdNode = etree_.SubElement(invocationInfoNode, msgTag('TransactionId'))
        invocationStateNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationState'))

        errorTexts = []

        operationHandleRefs = request.bodyNode.xpath('msg:SetContextState/msg:OperationHandleRef/text()',
                                                     namespaces=nsmap)
        if len(operationHandleRefs) == 1:
            operationHandleRef = operationHandleRefs[0]
            operation = self._sdcDevice.getOperationByHandle(operationHandleRef)
            if operation is None:
                errorTexts.append('operation "{}" not known'.format(operationHandleRef))
        elif len(operationHandleRefs) > 1:
            errorTexts.append('multiple OperationHandleRefs found: "{}"'.format(operationHandleRefs))
        else:
            errorTexts.append('no OperationHandleRef found')

        if errorTexts:
            invocationStateNode.text = pmtypes.InvocationState.FAILED
            transactionIdNode.text = '0'
            operationErrorNode = etree_.SubElement(invocationInfoNode, msgTag('InvocationError'))
            operationErrorNode.text = pmtypes.InvocationError.INVALID_VALUE
            operationErrorMessageNode = etree_.SubElement(invocationInfoNode,
                                                          msgTag('InvocationErrorMessage'))
            operationErrorMessageNode.text = '; '.join(errorTexts)
        else:
            transactionId = self._sdcDevice.enqueueOperation(operation, request)
            transactionIdNode.text = str(transactionId)
            invocationStateNode.text = pmtypes.InvocationState.WAIT

        response.addBodyElement(replyBodyNode)
        return response

    def _onGetContextStates(self, httpHeader, request):  # pylint:disable=unused-argument
        self._logger.debug('_onGetContextStates')
        requestedHandles = request.bodyNode.xpath('*/msg:HandleRef/text()', namespaces=nsmap)
        if len(requestedHandles) > 0:
            self._logger.info('_onGetContextStates requested Handles:{}', requestedHandles)
        nsmapper = self._mdib.nsmapper
        response = pysoap.soapenvelope.Soap12Envelope(
            nsmapper.partialMap(Prefix.S12, Prefix.WSA, Prefix.PM, Prefix.MSG))
        replyAddress = request.address.mkReplyAddress(action=self._getActionString('GetContextStatesResponse'))
        response.addHeaderObject(replyAddress)
        getContextStatesResponseNode = etree_.Element(msgTag('GetContextStatesResponse'))
        with self._mdib.mdibLock:
            self._mdib.mdib_version_group.update_node(getContextStatesResponseNode)
            if len(requestedHandles) == 0:
                # MessageModel: If the HANDLE reference list is empty, all states in the MDIB SHALL be included in the result list.
                contextStateContainers = list(self._mdib.contextStates.objects)
            else:
                contextStateContainersLookup = OrderedDict() # lookup to avoid double entries
                for handle in requestedHandles:
                    # If a HANDLE reference does match a multi state HANDLE,
                    # the corresponding multi state SHALL be included in the result list
                    tmp = self._mdib.contextStates.handle.getOne(handle, allowNone=True)
                    if tmp:
                        tmp = [tmp]
                    if not tmp:
                        # If a HANDLE reference does match a descriptor HANDLE,
                        # all states that belong to the corresponding descriptor SHALL be included in the result list
                        tmp = self._mdib.contextStates.descriptorHandle.get(handle)
                    if not tmp:
                        # R5042: If a HANDLE reference from the msg:GetContextStates/msg:HandleRef list does match an
                        # MDS descriptor, then all context states that are part of this MDS SHALL be included in the result list.
                        descr = self._mdib.descriptions.handle.getOne(handle, allowNone=True)
                        if descr:
                            if descr.NODETYPE == domTag('MdsDescriptor'):
                                tmp = list(self._mdib.contextStates.objects)
                    if tmp:
                        for st in tmp:
                            contextStateContainersLookup[st.Handle] = st
                contextStateContainers = contextStateContainersLookup.values()
            if contextStateContainers:
                for contextStateContainer in contextStateContainers:
                    node = contextStateContainer.mkStateNode()
                    getContextStatesResponseNode.append(node)
                    node.tag = msgTag('ContextState')
        response.addBodyElement(getContextStatesResponseNode)
        self._logger.debug('_onGetContextStates returns {}', lambda: response.as_xml(pretty=False))
        return response

    def addWsdlPortType(self, parentNode):
        portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                     attrib={'name': self.port_type_string,
                                             dpwsTag('DiscoveryType'):'dt:ServiceProvider',
                                             wseTag('EventSource'): 'true'})
        _mkWsdlTwowayOperation(portType, operationName='SetContextState')
        _mkWsdlTwowayOperation(portType, operationName='GetContextStates')
        _mkWsdlOnewayOperation(portType, operationName='EpisodicContextReport')
        _mkWsdlOnewayOperation(portType, operationName='PeriodicContextReport')



class DescriptionEventService(DPWSPortTypeImpl):
    WSDLMessageDescriptions = (WSDLMessageDescription('DescriptionModificationReport', ('{}:DescriptionModificationReport'.format(_msg_prefix),)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('DescriptionModificationReport', None, 'literal'),
                             )

    def addWsdlPortType(self, parentNode):
        portType = etree_.SubElement(parentNode, etree_.QName(_wsdl_ns,'portType'),
                                     attrib={'name':self.port_type_string,
                                             dpwsTag('DiscoveryType'):'dt:ServiceProvider',
                                             wseTag('EventSource'): 'true'})
        _mkWsdlOnewayOperation(portType, operationName='DescriptionModificationReport')





class RequestManipulator:
    """
    This class can be used to inspect or manipulate output created by the sdc client.
    Output creation is done in three steps: first a pysdc.pysoap.soapenvelope.Soap12Envelope is created,
    then a (libxml) etree is created from its content, anf finally a bytestring is generated from the etree.
    The sdc client calls corresponding methods of the manipulator object after every step.
    If the method returns something different from None, this returned value will be used as input for the next step.
    """
    def __init__(self, cb_soapenvelope=None, cb_xml=None, cb_string=None):
        """

        :param cb_soapenvelope: a callback that gets the SoapEnvelope instance as parameter
        :param cb_xml:  a callback that gets the etree instance as parameter
        :param cb_string:  a callback that gets the output string as parameter
        """
        self.cb_soapenvelope = cb_soapenvelope
        self.cb_xml = cb_xml
        self.cb_string = cb_string

    def manipulate_soapenvelope(self, soap_envelope):
        if callable(self.cb_soapenvelope):
            return self.cb_soapenvelope(soap_envelope)

    def manipulate_domtree(self, domtree):
        if callable(self.cb_xml):
            return self.cb_xml(domtree)

    def manipulate_string(self, xml_string):
        if callable(self.cb_string):
            return self.cb_string(xml_string)




from functools import partial
from .. import observableproperties as properties
from .. import loghelper
from .. import pmtypes
from .. import sdcdevice



class ProviderRole(object):
    def __init__(self, log_prefix):
        self._mdib = None
        self._logger = loghelper.getLoggerAdapter('sdc.device.{}'.format(self.__class__.__name__), log_prefix)

    def stop(self):
        """ if provider uses worker threads, implement stop method"""
        pass

    def initOperations(self, mdib):
        self._mdib = mdib

    def makeOperationInstance(self, operationDescriptorContainer): #pylint: disable=unused-argument
        """returns a callable for this operation or None.
        If a mdib already has operations defined, this method can connect a handler to a given operation descriptor.
        Use case: initialization from an existing mdib"""
        return None

    def makeMissingOperations(self):
        """
        This method is called after all existing operations from mdib have been registered.
        If a role provider needs to add operations beyond that, it can do it here.
        :return: []
        """
        return []

    def onPreCommit(self, mdib, transaction):
        pass

    def onPostCommit(self, mdib, transaction):
        pass


    def _setNumericValue(self, operationInstance, value):
        """ sets a numerical metric value"""
        operationDescriptorHandle = operationInstance.handle
        operationDescriptorContainer = self._mdib.descriptions.handle.getOne(operationDescriptorHandle)
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        self._logger.info('set value of {} via {} from {} to {}', operationTargetHandle, operationDescriptorHandle,
                          operationInstance.currentValue, value)
        operationInstance.currentValue = value
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getMetricState(operationTargetHandle)
            if state.metricValue is None:
                state.mkMetricValue()
            state.metricValue.Value = value
            #SF1823: For Metrics with the MetricCategory = Set|Preset that are being modified as a result of a
            # SetValue or SetString operation a Metric Provider shall set the MetricQuality / Validity = Vld.
            metricDescriptorContainer = self._mdib.descriptions.handle.getOne(operationTargetHandle)
            if metricDescriptorContainer.MetricCategory in (pmtypes.MetricCategory.SETTING,
                                                            pmtypes.MetricCategory.PRESETTING):
                state.metricValue.Validity = pmtypes.MeasurementValidity.VALID


    def _setString(self, operationInstance, value):
        """ sets a string value"""
        operationDescriptorHandle = operationInstance.handle
        operationDescriptorContainer = self._mdib.descriptions.handle.getOne(operationDescriptorHandle)
        operationTargetHandle = operationDescriptorContainer.OperationTarget
        self._logger.info('set value {} from {} to {}', operationTargetHandle, operationInstance.currentValue,
                          value)
        operationInstance.currentValue = value
        with self._mdib.mdibUpdateTransaction() as mgr:
            state = mgr.getMetricState(operationTargetHandle)
            if state.metricValue is None:
                state.mkMetricValue()
            state.metricValue.Value = value
            #SF1823: For Metrics with the MetricCategory = Set|Preset that are being modified as a result of a
            # SetValue or SetString operation a Metric Provider shall set the MetricQuality / Validity = Vld.
            metricDescriptorContainer = self._mdib.descriptions.handle.getOne(operationTargetHandle)
            if metricDescriptorContainer.MetricCategory in (pmtypes.MetricCategory.SETTING,
                                                            pmtypes.MetricCategory.PRESETTING):
                state.metricValue.Validity = pmtypes.MeasurementValidity.VALID


    def _mkOperationFromOperationDescriptor(self, operationDescriptorContainer,
                                            currentArgumentHandler=None, currentRequestHandler=None,
                                            timeoutHandler=None):
        """
        :param operationDescriptorContainer: the operation container for which this operation Handler shall be created
        :param currentArgumentHandler: the handler that shall be called by operation
        :param currentRequestHandler: the handler that shall be called by operation
        :param timeoutHandler: callable when timeout is detected (InvocationEffectiveTimeout)
        :return: instance of cls
        """
        cls = sdcdevice.sco.getOperationClass(operationDescriptorContainer.NODETYPE)
        op = self._mkOperation(cls,
                               operationDescriptorContainer.handle,
                               operationDescriptorContainer.OperationTarget,
                               operationDescriptorContainer.coding,
                               currentArgumentHandler,
                               currentRequestHandler,
                               timeoutHandler)
        op.safetyReq = operationDescriptorContainer.SafetyReq
        return op

    def _mkOperation(self, cls, handle, operationTargetHandle, codedValue,
                     currentArgumentHandler=None, currentRequestHandler=None, timeoutHandler=None):
        """

        :param cls: one of the Operations defined in sdcdevice.sco
        :param handle: the handle of this operation
        :param operationTargetHandle: the handle of the operation target
        :param codedValue: the CodedValue for the Operation ( can be None)
        :param currentArgumentHandler: the handler that shall be called by operation
        :param currentRequestHandler: the handler that shall be called by operation
        :return: instance of cls
        """
        operation = cls(handle=handle,
                        operationTarget=operationTargetHandle,
                        codedValue=codedValue)
        if currentArgumentHandler:
            # bind method to currentArgument
            properties.strongbind(operation, currentArgument=partial(currentArgumentHandler, operation))
        if currentRequestHandler:
            # bind method to currentRequest
            properties.strongbind(operation, currentRequest=partial(currentRequestHandler, operation))
        if timeoutHandler:
            # bind method to onTimeout
            properties.strongbind(operation, onTimeout=partial(timeoutHandler, operation))
        return operation